import numpy as np
import os
import pandas as pd

from datetime import datetime
from sklearn.model_selection import StratifiedKFold
from time import mktime
from timeit import default_timer as dt

from .distributions import (likelihood_bernoulli, likelihood_poisson,
                            likelihood_gaussian)
from .metrics import aicCalc
from .parallel import p_prog_simp
#from .tmath import percentIncrease, powerset
from .utils import updateProgBar, print_time


def naiveVarDrop(X, searchCols, tol=0.0001, standardize=False, asList=False,
                 print_=False):
    '''
    Drop columns based on which columns have variance below the threshold.

    Parameters
    ----------
    X : pandas dataframe
        Feature (or design) matrix.
    searchCols : list or list-like (str)
        The columns to search. If None, use all columns. Default is None.
    tol : float, optional
        The threshold for variance to decide which columns to keep or drop.
        The default is 0.0001.
    asList : bool, optional
        Return only the list of columns to be dropped. The default is False.
    print_ : bool, optional
        Print the columns to be dropped. The default is False.

    Returns
    -------
    list or dataframe
        Either list of columns to be dropped or dataframe with columns removed.

    '''
    cols_to_drop = []
    if searchCols is None:
        searchCols = list(X.columns)
    for i in searchCols:
        var = X.loc[:, i].var(ddof=1)
        if var < tol:
            cols_to_drop.append(i)
    if print_:
        print("Dropped " + str(len(cols_to_drop)) + " low-var Columns")
    if asList:
        return cols_to_drop
    else:
        return X.drop(cols_to_drop, axis=1)


def naiveScoreDrop(X, scores, tol=0.001, asList=False):
    '''
    Drop columns based on which columns have scores below the threshold. This 
    could be used with any arbitrary score function, where scores is a 
    1-column dataframe, in which the index are column names, and the values
    are the scores.

    Parameters
    ----------
    X : pandas dataframe
        Feature (or design) matrix.
    scores : pandas dataframe
        The score results for each column.
    tol : float, optional
        The threshold for variance to decide which columns to keep or drop.
        The default is 0.0001.
    asList : bool, optional
        Return only the list of columns to be dropped. The default is False.

    Returns
    -------
    list or dataframe
        Either list of columns to be dropped or dataframe with columns removed.
    '''
    cols_to_drop = []
    for i in scores.index:
        score = scores.loc[i].values
        if score < tol:
            cols_to_drop.append(i)
    if asList:
        return cols_to_drop
    else:
        return X.drop(cols_to_drop, axis=1)
    
    
def getHighCorrs(corr_mat, corr_thres, split_key="..&.."):
    '''
    Given a correlation matrix, return the
    correlations that are above the threshold.
    
    To be used before dropHighCorrs.

    Parameters
    ----------
    corr_mat : pandas dataframe
        The correlation matrix.
    corr_thres : float
        The threshold for correlation to decide which columns to keep or drop.
        Between 0 and 1.
    split_key : str, optional
        The unique key to use to join the column names together, for example,
        if split_key was '..&..', then the index for that correlation would be
        'col1..&..col2'. The default is "..&..".

    Returns
    -------
    top_corr : pandas dataframe
        A dataframe containing a list of the top correlations from the
        correlation matrix.

    '''
    top_corr = {}
    columns = corr_mat.columns
    for i in corr_mat.columns:
        for j in columns:
            if i == j:
                continue
            if np.isnan(corr_mat.loc[i, j]):
                continue
            cur_corr = corr_mat.loc[i, j]
            if abs(cur_corr) >= corr_thres:
                top_corr[split_key.join([i, j])] = cur_corr
        columns = columns.drop(i)
    return top_corr


def dropHighCorrs(X, top_corr, split_key="..&..", asList=False, print_=False):
    '''
    Remove high inter-correlations from the dataset. When dropping a column,
    the column with the lowest variance is selected.
    
    To be used after getHighCorrs.

    Parameters
    ----------
    X : pandas dataframe
        The data in tabular form (the feature or design matrix of which the 
        correlation is being evaluated).
    top_corr : dataframe
        The output from the getHighCorrs function, or, a dataframe containing a
        list of the high correlations, where the index is a list of the column
        names concatenated together by split_key, for example, 'col1..&..col2'.
    split_key : str, optional
        The unique key to use to join the column names together, for example,
        if split_key was '..&..', then the index for that correlation would be
        'col1..&..col2'. The default is "..&..".
    asList : bool, optional
        Return only the list of columns to be dropped. The default is False.
    print_ : bool, optional
        Print the columns to be dropped. The default is False.

    Returns
    -------
    list or dataframe
        Either list of columns to be dropped or dataframe with columns removed.

    '''
    cols_to_drop = []
    for i in top_corr.keys():
        cols = i.split(split_key)
        var1 = np.nanvar(X.loc[:, cols[0]])
        var2 = np.nanvar(X.loc[:, cols[1]])
        if var1 < var2:
            cols_to_drop.append(cols[0])
        else:
            cols_to_drop.append(cols[1])
    if print_:
        print("Dropped " + str(len(cols_to_drop)) + " high-corr Columns")
    if asList:
        return cols_to_drop
    else:
        return X.drop(cols_to_drop, axis=1)


def dropCorrProcedure(X, corr_thres, split_key="..&..", 
                      asList=False, print_=False):
    # TODO: add an option for dropping columns that have a low correlation with
    # the target
    '''
    Calculates the inter-correlation of the columns in a matrix/dataframe, and then
    drops columns that are above a given threshold.

    Parameters
    ----------
    X : numpy array or pandas dataframe
        The data in tabular form (the feature or design matrix). If not a 
        dataframe, it will be converted.
    corr_thres : float
        The threshold for correlation to decide which columns to keep or drop.
        Between 0 and 1.
    split_key : str, optional
        The unique key to use to join the column names together, for example,
        if split_key was '..&..', then the index for that correlation would be
        'col1..&..col2'. The default is "..&..".
    asList : bool, optional
        Return only the list of columns to be dropped. The default is False.
    print_ : bool, optional
        Print the columns to be dropped. The default is False.

    Returns
    -------
    dropped_cor : dataframe
        Dataframe with High correlation columns dropped.

    '''
    if not isinstance(X, pd.DataFrame):
        X = pd.DataFrame(X.copy())
    corr_mat = X.corr()
    top_corr = getHighCorrs(corr_mat, corr_thres, split_key=split_key)
    dropped_cor = dropHighCorrs(X, top_corr, split_key=split_key,
                                asList=asList, print_=print_)
    return dropped_cor


def permutation_importance(fit_model, Xtest, Ytest,
                           metric_func, seed=None, sort="dsc"):
    '''
    Performs permutation importance on a fitted model.
    
    Assumes that your metric function takes inputs in this order: ytrue, yperd.
    If it doesn't, write a simple wrapper that will. Also assumes a pretrained
    model with a predict method.
    
    Credit: https://explained.ai/rf-importance/
    
    This function was adapted from the above link.

    Parameters
    ----------
    fit_model : sklearn, (or similar)
        Can be any model that has a 'predict' method.
    Xtest : pandas Dataframe
        The observations in the feature or design matrix to perform the
        permutation test on.
    Ytest : pandas Series
        The observations in the response variable to perform the
        permutation test on.
    metric_func : function
        The function to evaluate your metric of interest.
    seed : int, optional
        Set Random Seed for reporducibility. The default is None.
    sort : str, optional
        Either 'asc' for ascending or 'dsc' for descending.
        The default is "dsc".

    Returns
    -------
    res : dataframe
        A dataframe containing the results of the permutation test.

    '''
    Ypred = fit_model.predict(Xtest)
    baseline = metric_func(Ytest, Ypred)
    imps = []
    if seed is not None:
        np.random.seed(seed)
    for col in Xtest.columns:
        col_copy = Xtest.loc[:, col].copy()
        Xtest.loc[:, col] = np.random.permutation(Xtest.loc[:, col])
        Ypred = fit_model.predict(Xtest)
        score = metric_func(Ytest, Ypred)
        Xtest.loc[:, col] = col_copy
        imps.append(baseline - score)
    res = pd.DataFrame(imps, index=Xtest.columns, columns=["Importance"])
    # Could reduce the sorting to a one-liner, but this way someone
    # could choose not to sort. Though, probably not a useful feature.
    if sort == "asc":
        res = res.sort_values('Importance', ascending=True)
    elif sort == "dsc":
        res = res.sort_values('Importance', ascending=False)
    return res


def permutation_importance_CV(model, X, Y, metric_func, num_splits=8, seed=None, sort="dsc"):
    '''
    Performs a cross-validated permutation importance on a non-fitted model.
    
    Assumes that your metric function takes inputs in this order: ytrue, yperd.
    If it doesn't, write a simple wrapper that will. Also assumes a pretrained
    model with a predict method.
    
    Credit: https://explained.ai/rf-importance/
    
    This function was adapted from the above link.

    Parameters
    ----------
    model : sklearn, (or similar)
        Can be any model that has a 'predict' method. Send to the function 
        unfitted, only instantiated
    X : pandas Dataframe
        The observations in the feature or design matrix to perform the
        permutation test on.
    Y : pandas Series
        The observations in the response variable to perform the
        permutation test on.
    metric_func : function
        The function to evaluate your metric of interest.
    num_splits : int
        Number of CV folds to use (using stratified k-fold)
    seed : int, optional
        Set Random Seed for reporducibility. The default is None.
    sort : str, optional
        Either 'asc' for ascending or 'dsc' for descending.
        The default is "dsc".

    Returns
    -------
    res : dataframe
        A dataframe containing the results of the permutation test.

    '''
    splits = StratifiedKFold(n_splits=num_splits, shuffle=True, random_state=seed)
    pi_total = np.zeros((X.shape[1], 1))
    i = 0
    for train_idx, test_idx in splits.split(X, Y):
        Xtrain, Xtest = X.iloc[train_idx], X.iloc[test_idx]
        Ytrain, Ytest = Y.iloc[train_idx], Y.iloc[test_idx]
        fi_model = model.fit(Xtrain, Ytrain)
        pi = permutation_importance(fi_model, Xtest, Ytest,
                                metric_func, seed=seed, sort="None")
        pi_total += pi.values
        i += 1
        print("Completed " + str(i) + " of " + str(num_splits))
    pi_total /= num_splits
    pi_tot_df = pd.DataFrame(pi_total, index=X.columns, columns=["Importance"])
    # Could reduce the sorting to a one-liner, but this way someone
    # could choose not to sort. Though, probably not a useful feature.
    if sort == "asc":
        pi_tot_df = pi_tot_df.sort_values('Importance', ascending=True)
    elif sort == "dsc":
        pi_tot_df = pi_tot_df.sort_values('Importance', ascending=False)
    return pi_tot_df


def _calcForwardAICs(X, Y, model, metric, family):
    '''
    A utility function for the forwardSelection algorithm. Calculates AIC
    values for a two-class classification model.

    Parameters
    ----------
    X : pandas Dataframe
        The feature or design matrix.
    Y : pandas Series
        The response variable.
    model : sklearn, or similar
        Any model that has a fit and predict (or predict_proba) method.
    metric : str
        The AIC metric to be used, for example, aic, aicc, bic, ebic, hastie, 
        or kwano.
    family : str
        The family of distributions to calculate log-likelihood from.

    Returns
    -------
    score : float
        The AIC (or variant) score.

    '''
    mod = model.fit(X, Y)

    ss = X.shape[0]
    ncoefs = X.shape[1]

    if family == "binomial":
        nllf = likelihood_bernoulli
        # Assumes the probability of positive class
        y_pred = mod.predict_proba(X)[:, 1]
    elif family == "poisson":
        nllf = likelihood_poisson
        y_pred = mod.predict(X)
    elif family == "gaussian":
        nllf = likelihood_gaussian
        y_pred = mod.predict(X)
    else:
        raise ValueError("Not a valid family")

    loglike = nllf(y_true=Y, y_score=y_pred, neg=False)

    score = aicCalc(loglike, ncoefs, sample_size=ss, c=2, metric=metric)
    return score


def _doParallelForward(save_path, save_name, save_ext, metric, model,
                      target_var, new_predictor, use_probabilities,
                      mod_cols, family, Yprob=None,
                      serialize_flavor='feather'):
    '''
    A helper function to lessen the load on memory when computing in parallel.

    Parameters
    ----------
    save_path : str
        Path to temporary saved data.
    save_name : str
        Name of temporary data.
    save_ext : str
        The name of the temporary file extension.
    metric : str
        The AIC metric to be used, for example, aic, aicc, bic, ebic, hastie, 
        or kwano.
    model : sklearn, or similar
        Any model that has a fit and predict method.
    target_var : str
        The column containing the target (or response) variable.
    new_predictor : str
        The column name of the most recently added predictor.
    use_probabilities : bool
        Whether or not to use predicted probabilities as the only other feature
        in the set, or to use the actual features in performing the forward 
        selection.
    mod_cols : list
        Columns that made it into the final model.
    family : str
        The family of distributions to calculate log-likelihood from.
    Yprob : numpy array or pandas series, optional
        The predicted probabilities from the current model, if applicable.
        The default is None.
    serialize_flavor : str
        Which mode of downsaving data to use, currently supports 'feather' and
        'msgpack'. Unfortunately, as of pandas 0.25.0, msgpack is no longer 
        supported.
        
        Note: using feather requires pyarrow

    Returns
    -------
    res : float
        The AIC (or variant) result.

    '''
    
    # Note: Should probably update this to feather (as of 2019)
    # since it is faster and better on memory in general.
    # Since this is not long term storage, feather would be a 
    # good choice. However, for long-term storage, parquet is 
    # probably the best option

    if serialize_flavor == 'msgpack':
        data = pd.read_msgpack(save_path + save_name + save_ext)
        chunks = [int(x[6:]) for x in data.keys()]
        chunks = sorted(chunks)
        mergedDF = pd.DataFrame()
        for chunk in chunks:
            mergedDF = mergedDF.append(data['chunk_'+str(chunk)])
    elif serialize_flavor == 'feather':
        mergedDF = pd.read_feather(save_path + save_name + save_ext)
    else:
        raise ValueError("{0} is not a supported serialize method.".format(serialize_flavor))
    

    Y = mergedDF.loc[:, target_var]

    if use_probabilities:
        X = pd.DataFrame({'Yprob': Yprob,
                          new_predictor: mergedDF.loc[:, new_predictor].values
                         }, index=mergedDF.index)
    else:
        X = mergedDF.loc[:, mod_cols + new_predictor]

    res = _calcForwardAICs(X, Y, model, metric, family)

    # For debugging purposes, to attempt to force garbage collection in parallel
    #data = []
    #chunks = []
    #mergedDF = []
    #X = []
    #Y = []
        
    return res

def _prepare_for_parallel(XY, serialize_flavor='feather'):
    '''
    Utility function to setup the data for parallel processing with low
    memory overhead.

    Parameters
    ----------
    XY : pandas dataframe
        The combined independent variables/features and reponse/target
        variable.

    Returns
    -------
    save_name : str
        The name of the temporary data.
    save_path : str
        The name of the temporary file path.
    save_ext : str
        The name of the temporary file extension.
    num_chunks : int
        The number of chunks inside pandas msgpck format (determined by 
        dataframe size).
    serialize_flavor : str
        Which mode of downsaving data to use, currently supports 'feather' and
        'msgpack'. Unfortunately, as of pandas 0.25.0, msgpack is no longer 
        supported.
        
        Note: using feather requires pyarrow

    '''
        
    save_path = 'forward_tmp'
    if not os.path.isdir(save_path):
        os.mkdir(save_path)
    save_name = '/forward_run_' + str(int(mktime(datetime.now().timetuple())))
    save_ext = ''
    
    if serialize_flavor == 'msgpack':
        save_ext = '.msg'
        num_chunks = int(((XY.memory_usage(index=True).sum()/(1024**3))/2) + 1)
        pd.to_msgpack(save_path + save_name + save_ext, {
            'chunk_{0}'.format(i):chunk for i, chunk in enumerate(np.array_split(XY, num_chunks))
            })
    elif serialize_flavor == 'feather':
        save_ext = '.fth'
        num_chunks = None
        XY.to_feather(save_path + save_name + save_ext)
    else:
        raise ValueError("{0} is not a supported serialize method.".format(serialize_flavor))
    
    return save_name, save_path, save_ext, num_chunks


def forwardSelection(XY, target_var, model, metric='bic', family='gaussian',
                     verbose=True, n_jobs=1, early_stop=False, perc_min=0.01,
                     stop_at_p=1000, stop_when=5, use_probabilities=False,
                     return_type='all', serialze_flavor='feather',
                     use_threads=True):
    '''
    A forward selection algorithm for classification only right now. Still 
    needs some work.
    
    Note: it is a known bug for this function to fail when n_jobs > 1 while
    using sypder. This is because of an issue with spyder and the p_prog_simp
    function. I have not been able to discover why, or provide a fix.
    If you run a script with this function from the console, it should run
    just fine. 

    Parameters
    ----------
    XY : pandas dataframe
        The combined independent variables/features and reponse/target
        variable.
    target_var : str
        The column containing the target (or response) variable.
    model : sklearn, or similar
        An unfitted model. Any model that has a fit and predict method.
    metric : str
        The AIC metric to be used, for example, aic, aicc, bic, ebic, hastie, 
        or kwano.
    family : str
        The family of distributions to calculate log-likelihood from.
    verbose : bool, optional
        Output the steps and progress as it completes. The default is True.
    n_jobs : int, optional
        If greater than 1, perform operation in parallel. The default is 1.
    perc_min : float, optional
        --NOT IMPLEMENTED--. The percent change minimum for breaking early,
        if no meanigful change in metric is detected. The default is 0.01.
    stop_at_p : int, optional
        The number of selections to stop at. In other words, it selects up to
        `p` predictors, even if the optimal model is not yet found and more
        tests could be done. The default is 1000.
    stop_when : int, optional
        Stop the operations when the metric no longer continues to decrease,
        after evaluating the next stop_when columns. The default is 5.
    use_probabilities : bool, optional
        Whether or not to use predicted probabilities as the only other feature
        in the set, or to use the actual features in performing the forward 
        selection. The default is False.
    return_type : str, optional
        Which object to return, can be either list, model, data, or all.
        The default is 'all'.
    serialize_flavor : str
        Which mode of downsaving data to use, currently supports 'feather' and
        'msgpack'. Unfortunately, as of pandas 0.25.0, msgpack is no longer 
        supported.
    use_threads : bool, optional
        Use threads instead of processes for parallel operation.f
        
        Note: using feather requires pyarrow

    Returns
    -------
    list, model, data, or tuple of objects
        Either return as list, model, data, or all. The default is 'all'.

    '''
    
    t0 = dt()
    
    total_possible_complexity = int(((XY.shape[1]-1)**2) - ((XY.shape[1]-1)*((XY.shape[1]-1) - 1)/2))
    
    if early_stop:
        if stop_at_p >= (XY.shape[1] - 1):
            complexity = total_possible_complexity
            stop_at_p = XY.shape[1] - 1
        else:
            complexity = np.sum((XY.shape[1] - 1) - np.arange(0, stop_at_p))
    else:
        complexity = total_possible_complexity
        stop_at_p = XY.shape[1] - 1
    
    
    if verbose:
        print_time("Problem Complexity: " + str(complexity) + " Iterations Needed, " +
                   str(total_possible_complexity) + " Possible...",
                   t0, te=dt(), backsn=True)
    
    # Use this for an exhaustive search
    #all_combos = list(powerset(X.columns))
    final_mod_cols = []
    leftover_cols = [x for x in XY.columns if x != target_var]
    currentScore = np.inf
    #early_stop_counter = 0
    loop_counter = 0
    
    if n_jobs > 1:
        if verbose:
            print_time("\nPreparing Parallel Operation...", t0, te=dt())
            
        save_name, save_path, save_ext, num_chunks = _prepare_for_parallel(XY)
        arg = {'save_path': save_path,
               'save_name': save_name,
               'save_ext': save_ext,
               'metric': metric,
               'model': model,
               'target_var': target_var,
               'family': family,
               'use_probabilities': use_probabilities,
               'mod_cols': final_mod_cols,
               'Yprob': None
              }
    
    for i in range(XY.shape[1] - 1):
        scores = []
        if use_probabilities and i > 1:
            # create current model and probability array
            initial_fit = model.fit(XY.loc[:, final_mod_cols], XY.loc[:, target_var])
            Yprob = initial_fit.predict_proba(XY.loc[:, final_mod_cols])[:, 1]
            #print('\nprob gen step: ', Yprob.shape, "\n")
            if n_jobs > 1:
                arg['Yprob'] = Yprob
        
        if n_jobs > 1:
            if verbose:
                print_time("\nPerforming " + str(i + 1) + " of " + str(stop_at_p) + " Steps...",
                           t0, te=dt(), backsn=True)
            loop_arg = [{'new_predictor': [col]} for col in leftover_cols]
            scores = p_prog_simp(arg, loop_arg, _doParallelForward, n_jobs,
                                 use_threads=use_threads)
            #return scores
            if verbose:
                loop_counter += len(loop_arg)
                updateProgBar(loop_counter, complexity, t0)
        else:
            for k, col in enumerate(leftover_cols):
                if use_probabilities and i > 1:
                    #print('loop step: ', XY.loc[:, col].values.shape, "\n")
                    XY_prob = pd.DataFrame({'Yprob': Yprob, col: XY.loc[:, col].values}, index=XY.index)
                    score = _calcForwardAICs(XY_prob, XY.loc[:, target_var],
                                             model, metric, family)
                else:
                    score = _calcForwardAICs(XY.loc[:, final_mod_cols + [col]], XY.loc[:, target_var],
                                             model, metric, family)
                scores.append(score)
            
                if verbose:
                    loop_counter += 1
                    updateProgBar(loop_counter, complexity, t0)
        
        minScore = np.min(scores)
        if minScore > currentScore:
            print("\nNo further imporovement in the model. Breaking...")
            break
        else:
            currentScore = minScore
        
        minScoreLoc = np.where(minScore == np.array(scores))[0][0]
        final_mod_cols.append(leftover_cols[minScoreLoc])
        del leftover_cols[minScoreLoc]
        
        # could be useful if I wanted to do something like accuracy,
        # but for AIC/BIC, this isn't necessary
        #if np.abs(percentIncrease(minScore, prev_minScore)) < perc_min:
        #    early_stop_counter += 1
        #else:
        #    early_stop_counter = 0
        #
        #if early_stop_counter >= stop_when:
        #    print('Stopped Early')
        #    break
        #if i >= stop_at_p:
        #    print('Stopped at ', str(i), ' selected predictors')
        #    break

    if n_jobs > 1:
        file_ = save_path + save_name + save_ext
        try:
            os.remove(file_)
        except FileNotFoundError:
            print('Could Not find file ', file_, 'Continuing...')
            
    if return_type == 'list':
        return final_mod_cols
    elif return_type == 'model':
        return model.fit(XY.loc[:, final_mod_cols], XY.loc[:, target_var])
    elif return_type == 'data':
        return XY.loc[:, final_mod_cols]
    else:
        return final_mod_cols, model.fit(XY.loc[:, final_mod_cols],
                                         XY.loc[:, target_var]), XY.loc[:, final_mod_cols]

## TODO: import vif from modeling and incorporate a vif drop  
def vifDrop():
    return 0



import os
from . import distributions
from . import estimators
from . import feature_selection
from . import mcmc
from . import metrics
from . import modeling
from . import optimization
from . import parallel
from . import quick_analysis
from . import sampling
from . import tmath
from . import utils
from . import nn

__version__ = '1.0.10'

__all__ = ['distributions',
           'estimators',
           'feature_selection',
           'mcmc',
           'metrics',
           'modeling',
	   'optimization',
           'parallel',
           'quick_analysis',
           'sampling',
           'tmath',
           'utils',
	   'nn']



import numpy as np
import pandas as pd
import warnings

from copy import deepcopy as copy
from sklearn.metrics import (f1_score, confusion_matrix,
                             roc_auc_score, accuracy_score)
from sklearn.model_selection import (StratifiedKFold, KFold,
                                     ShuffleSplit, StratifiedShuffleSplit)
from timeit import default_timer as dt

from .metrics import (cox_snell_r2, nagelkerke_r2, tjur_r2, mcfadden_r2,
                      conf_mat_metrics, bias, rpmse, r2, adj_r2, top_20p,
                      number_of_nonzero_coef)
from .utils import reshape_to_vect


def scoreModel(X, Y, model, metrics=['Accuracy',
                                    'F1',
                                    'Sens/Recall',
                                    'Specificity',
                                    'ppv',
                                    'npv',
                                    'AUC'
                                    ],
               thres=None, mtype='classification',
               print_=True, avg="weighted"):
    '''
    Score a model using the given metrics.

    Parameters
    ----------
    X : pandas dataframe
        Feature or design marix.
    Y : pandas series
        Response or Target variable.
    model : sklearn, or similar
        model to fit to the data. Must have fit, predict, and/or predict_proba
        methods, depending on the metrics.
    metrics : list, optional
        The metrics to include in the analysis.
        The default is ['Accuracy', 'F1', 'Sens/Recall',
                        'Specificity', 'ppv', 'npv', 'AUC'].
    thres : float, optional
        The threshold to use for classification metrics where it might apply.
        If threshold is None, the default predict mehtod is used (which
        for most models uses a boundary at or equivalent to 0.5 probability).
        The default is None.
    mtype : str, optional
        The type of model, currently either classification or regressor.
        The default is 'classification'.
    print_ : bool, optional
        Print progress as applicable. The default is True.
    avg : str, optional
        for metrics where it applies, such as AUC, how to calculate the
        metric (see sklearn docs for more details, particularily with AUC).
        The default is "weighted".

    Returns
    -------
    res : dict
        A dictionary containing the results for each metric.

    '''
    res = {}
    args = {'average': avg,
            'conf_metric': None,
            'X': X,
            'y_true': Y,
            'y_pred': None,
            'y_score': None,
            'll_est': None,
            'll_null': None, 
            'y_prob_nullmod': None,
            'n_obs': None,
            'rsquared': None,
            'cs_r2': None,
            'model': model}
    All_Metrics = {
            'Accuracy': {
                         'Function': accuracy_score,
                         'arguments': ['y_true', 'y_pred']
                        },
            'F1': {
                   'Function': f1_score,
                   'arguments': ['y_true', 'y_pred']
                  },
            'Sens/Recall': {
                            'Function': conf_mat_metrics,
                            'arguments': ['y_true', 'y_pred', 'conf_metric']
                           },
            'Specificity': {
                            'Function': conf_mat_metrics,
                            'arguments': ['y_true', 'y_pred', 'conf_metric']
                           },
            'ppv': {
                    'Function': conf_mat_metrics,
                    'arguments': ['y_true', 'y_pred', 'conf_metric']
                   },
            'npv': {
                    'Function': conf_mat_metrics,
                    'arguments': ['y_true', 'y_pred', 'conf_metric']
                   },
            'Conf_Mat': {
                         'Function': conf_mat_metrics,
                         'arguments': ['y_true', 'y_pred', 'conf_metric']
                        },
            'AUC': {
                    'Function': roc_auc_score,
                    'arguments': ['y_true', 'y_score', 'average']
                        }, 
            'Cox/Snell R2': {
                            'Function': cox_snell_r2,
                            'arguments': ['ll_est', 'll_null', 'n_obs']
                        },
            'McFadden R2': {
                            'Function': mcfadden_r2,
                            'arguments': ['ll_est', 'll_null']
                        },
            'Tjur R2': {
                        'Function': tjur_r2,
                        'arguments': ['y_true', 'y_score']
                       },
            'Nagelkerke R2': {
                              'Function': nagelkerke_r2,
                              'arguments': ['ll_null', 'n_obs', 'cs_r2', 'll_est']
                             },
            'Bias': {
                             'Function': bias,
                             'arguments': ['y_true', 'y_pred']
                            },
            'RPMSE': {
                             'Function': rpmse,
                             'arguments': ['y_true', 'y_pred']
                            },
            'R2': {
                             'Function': r2,
                             'arguments': ['y_true', 'y_pred']
                            },
            'Adj. R2': {
                             'Function': adj_r2,
                             'arguments': ['y_true', 'y_pred', 'X', 'rsquared']
                            },
            'Top 20%': {
                        'Function': top_20p,
                        'arguments': ['y_true', 'y_score']
            },
            'Avg Number of Used Features': {
                        'Function': number_of_nonzero_coef,
                        'arguments': ['X', 'model']
            }
        }
    
    y_temp = Y.values.ravel()
    if mtype == "classification":
        
        args['n_obs'] = Y.shape[0]
        args['y_score'] = model.predict_proba(X)[:, 1]
        
        if thres is None:
            args['y_pred'] = model.predict(X)
        else:
            args['y_pred'] = np.sum([args['y_score'] >= thres], axis=0)
        
        if any((met in ['McFadden R2',
                        'Nagelkerke R2',
                        'Cox/Snell R2'] for met in metrics)):
            
            args['y_prob_nullmod'] = np.repeat(np.mean(y_temp), args['n_obs'])
            args['ll_est'] = np.sum(y_temp*np.log(args['y_score']) + (1 - y_temp)*np.log(1 - args['y_score']))
            args['ll_null'] = np.sum(y_temp*np.log(args['y_prob_nullmod']) + (1 - y_temp)*np.log(1 - args['y_prob_nullmod']))
                  
    else:
        if isinstance(Y, pd.DataFrame) or isinstance(Y, pd.Series):
            args['y_true'] = y_temp
        args['y_pred'] = model.predict(X)
    
    for met in metrics:
        args['conf_metric'] = met
        temp_args = {k: args[k] for k in All_Metrics[met]['arguments']}
        res[met] = All_Metrics[met]['Function'](**temp_args) 
    
    if print_:
        print(res)
    
    return res


# Currently, this takes just as long as scoreModel and the logic is simpler... keeping for legacy reasons
def scoreModel_legacy(X, Y, model, thres=None, mtype='classification', print_=True, avg="weighted"):
    res = None
    if mtype == "classification":
        nobs = Y.shape[0]
        Yprob = model.predict_proba(X)[:, 1]
        Yprob_nullmod = np.repeat(np.mean(Y.values), nobs)
        
        if thres is None:
            Ypred = model.predict(X)
        else:
            Ypred = np.sum([Yprob >= thres], axis=0)
        
        accs = accuracy_score(Y, Ypred)
        f1s = f1_score(Y, Ypred)
        aucs = roc_auc_score(Y, Yprob, average=avg)
        
        ll_est = np.sum(Y.values*np.log(Yprob) + (1 - Y.values)*np.log(1 - Yprob))
        ll_null = np.sum(Y.values*np.log(Yprob_nullmod) + (1 - Y.values)*np.log(1 - Yprob_nullmod))
        
        # Cox Snell R2
        ratio = 2/nobs
        cs_r2 = 1 - np.exp(ratio*(ll_null - ll_est))
        r2_max = 1 - np.exp(ratio*ll_null)

        # Correction to CS R2 to bound between 0, 1
        n_r2 = cs_r2/r2_max
        
        # When the saturated model is not available (common), then
        # Mcfaddens R2 is equivalent to the likelihood ratio r2
        m_r2 = 1 - (ll_est/ll_null)
        
        #Tjur R2
        y_mu1 = Yprob[Y == 1].mean()
        y_mu0 = Yprob[Y == 0].mean()
        t_r2 = np.abs(y_mu1 - y_mu0)

        # Assumes 1 as the positive class
        confMat = confusion_matrix(Y, Ypred)
        tn = confMat[0, 0]
        tp = confMat[1, 1]
        fn = confMat[0, 1]
        fp = confMat[1, 0]
        tnfp = tn + fp
        tpfn = tp + fn
        tpfp = tp + fp
        tnfn = tn + fn
        if tnfp != 0:
            spec = tn/tnfp
        else:
            spec = 0
        if tpfn != 0:
            sens = tp/tpfn
        else:
            sens = 0
        if tpfp != 0:
            ppv = tp/tpfp
        else:
            ppv = 0
        if tnfn != 0:
            npv = tn/tnfn
        else:
            npv = 0
            
        if print_:
            print("Accuracy       : ", accs)
            print("F1             : ", f1s)
            print("Sens/Recall    : ", sens)
            print("Specificty     : ", spec)
            print("Pos Prev Val   : ", ppv)
            print("Neg Prev Val   : ", npv)
            print("AUC            : ", aucs)
            print("R2 (McFadden)  : ", m_r2)
            print("R2 (Tjur)      : ", t_r2)
            print("R2 (Nagelkerke): ", n_r2)
            print("R2 (Cox/Snell) : ", cs_r2)
    
        res = {
            'Accuracy': accs,
            'F1': f1s,
            'Sens/Recall': sens,
            'Specificity': spec,
            'ppv': ppv,
            'npv': npv,
            'Conf_Mat': confMat,
            'AUC': aucs,
            'McFadden R2': m_r2,
            'Tjur R2': t_r2,
            'Nagelkerke R2': n_r2,
            'Cox/Snell R2': cs_r2
        }
                  
    else:
        Ypred = model.predict(X)
        Ydiff = Ypred - Y.values
        Ybar = np.mean(Y.values)
        
        bias = np.mean(Ydiff)
        rpmse = np.sqrt(np.mean(Ydiff**2))
        SST = np.sum((Y.values - Ybar)**2)
        #SSE = np.sum((Y.values - Ypred)**2)
        SSR = np.sum((Ypred - Ybar)**2)
        # also equal to 1 - SSE/SST
        # alternative R2: np.corrcoef((ypred, ytrue))**2
        r2 = SSR/SST
        adj_r2 = 1 - (1 - r2)*((X.shape[0] - 1)/(X.shape[0] - X.shape[1] - 1))
        
        if print_:
            print("Bias       : ", bias)
            print("RPMSE      : ", rpmse)
            print("R2         : ", r2)
            print("Adj. R2    : ", adj_r2)
    
        res = {
            'Bias': bias,
            'RPMSE': rpmse,
            'R2': r2,
            'Adj. R2': adj_r2
        }
        
    return res


def runScorers(X, Y, splits, model, mtype, metrics=['Accuracy',
                                                    'F1',
                                                    'Sens/Recall',
                                                    'Specificity',
                                                    'ppv',
                                                    'npv',
                                                    'AUC'],
               avg='weighted', calculate=['Out of Sample'],
               method_on_X=None, mox_args={}, Y_for_test_only=None,
               sample_limit=20):
    '''
    Performs the cross-validation.

    Parameters
    ----------
    X : pandas dataframe
        Feature or design marix.
    Y : pandas series
        Response or Target variable.
    splits : sklearn.model_selection object
        Contains the split type and information about the split.
    model : sklearn, or similar
        model to fit to the data. Must have fit, predict, and/or predict_proba
        methods, depending on the metrics.
    mtype : str, optional
        The type of model, currently either classification or regressor.
        The default is 'classification'.
    metrics : list, optional
        The metrics to include in the analysis.
        The default is ['Accuracy', 'F1', 'Sens/Recall',
                        'Specificity', 'ppv', 'npv', 'AUC'].
    avg : str, optional
        for metrics where it applies, such as AUC, how to calculate the
        metric (see sklearn docs for more details, particularily with AUC).
        The default is "weighted".
    calculate : list or str, optional
        What to calculate the metrics on, either in-sample or out-of-sample,
        or both. In-sample measures the metrics on the train set while
        out-of-sample measures on the test set.
        The default is ['Out of Sample'].
    method_on_X : function, optional
        An optional function to pass that performs an operation on X.
        For example, you could pass a function to perform PCA before
        performing the fits. This would enable the function to be applied to
        each level of split. The method used to transform/modify X can be
        of any class, the only requirement is that it has both fit and
        transform methods, and that the arguments for fit method accept both
        an X and Y argument, even if the Y argument does nothing. You may need
        to create a simple wrapper for this, if you have a known method you 
        want to use, but doesn't quite fit what you need. The default is None.
    mox_args : dict, optional
        Any optional arguments that get passed to the constructor of the
        method_on_X. The default is {}.
    Y_for_test_only : pandas series or numpy array, optional
        An alternate target variable to test on, mainly for the use case of
        training on one response (perhaps one that is more informative or
        restrictive), but then predicting on a seperate test set.
        The default is None.
    sample_limit : int, optional
        The minimum acceptable samples in a split for the response
        variable. Only applies to Y_for_test_only currently.
        The default is 20.

    Returns
    -------
    keys : dict
        A dictionary containing the results.

    '''
    #is_score = None
    #oos_score = None
    keys = {}
 
    for train_idx, test_idx in splits.split(X, Y):
        Xtrain, Xtest = X.iloc[train_idx], X.iloc[test_idx]
        Ytrain, Ytest = Y.iloc[train_idx], Y.iloc[test_idx]
        # Adding a note so I don't forget what this is for... 
        # This is for the situation where you want to train on one definition of truth
        # but test on another, for example, train on the definition where a positive case
        # is either a revoke or a deny, but test on the definition where a positive case
        # is only revoked 
        if Y_for_test_only is not None:
            new_Ytest = Y_for_test_only.iloc[test_idx]
            if mtype == 'classification':
                classes = np.unique(new_Ytest)
                counts = np.array([new_Ytest[new_Ytest.values == i].shape[0] for i in classes])
                if np.all(counts >= sample_limit):
                    Ytest = new_Ytest
                else:
                    warnings.warn("Not all classes have at least %d samples. Using default test set" % (sample_limit))
            else:
                Ytest = new_Ytest
        
        # Method on X must return a dataframe in your method
        # on X function, set Y=None as a default argument
        # and let it pass through if not needed
        if method_on_X is not None:
            method_fit = method_on_X(**mox_args).fit(Xtrain, Ytrain)
            Xtrain = method_fit.transform(Xtrain)
            Xtest = method_fit.transform(Xtest)
        
        mod = copy(model).fit(Xtrain, Ytrain.values.ravel())
        
        if len(keys) == 0:
            for sk in calculate:
                keys[sk] = {}
                if sk == 'Out of Sample':
                    score = scoreModel(Xtest, Ytest, mod, metrics=metrics,
                                       mtype=mtype, print_=False, avg=avg)
                elif sk == 'In Sample':
                    score = scoreModel(Xtrain, Ytrain, mod, metrics=metrics,
                                       mtype=mtype, print_=False, avg=avg)
                    
                for key in score.keys():
                    keys[sk][key] = [score[key]]
        else:
            for sk in calculate:
                if sk == 'Out of Sample':
                    score = scoreModel(Xtest, Ytest, mod, metrics=metrics,
                                       mtype=mtype, print_=False, avg=avg)
                elif sk == 'In Sample':
                    score = scoreModel(Xtrain, Ytrain, mod, metrics=metrics,
                                       mtype=mtype, print_=False, avg=avg)
                    
                for key in score.keys():
                    keys[sk][key].append(score[key])
    return keys


def printScores(scores):
    '''
    A helper function to print the outputs of the crossVal function.

    Currently, it does not print confusion matrices.    

    Parameters
    ----------
    scores : dict
        The dictionary of scores.

    Returns
    -------
    None.

    '''
    longest_key = 0
    for sample in scores.keys():
        for score_type in scores[sample].keys():
            if len(score_type) > longest_key:
                longest_key = len(score_type)
    for sample in scores.keys():
        print(sample, ": ", sep='')
        for score_type in scores[sample].keys():
            values = scores[sample][score_type]
            if score_type == 'Conf_Mat':
                continue
            else:
                spacer = longest_key - len(score_type)
                print('\t', score_type, ' '*spacer, ': ', '\u03BC: ', '{:.4f}'.format(np.mean(values)), ', ',
                      '\u03C3: ', '{:.4f}'.format(np.std(values, ddof=1)), ', ',
                      '2\u03C3 Interval (', '{:.4f}'.format(np.mean(values)-(2*np.std(values, ddof=1))), ', ',
                      '{:.4f}'.format(np.mean(values)+(2*np.std(values, ddof=1))), ")",
                      sep='')
        print('')
    return None

        
def crossVal(X, Y, cv_iterations, model, method='k-fold',
             mtype='classification', stratified=True, print_=True,
             random_state=None, method_on_X=None, mox_args={},
             avg='weighted', shuffle=True, test_size=0.1,
             calculate=['Out of Sample'],
             metrics=['Accuracy', 'F1', 'Sens/Recall', 'Specificity',
                      'ppv', 'npv', 'AUC'],
             Y_for_test_only=None, sample_limit=20):
    # TODO: add parallel support
    '''
    A custom cross-validation strategy. While sklearn has a lot of beefed up
    functionality regarding cross-validation strategies, I like this one
    because it puts most of the strategies I like to use in one place. Also,
    it allows for you to output and calculate several cross-validation metrics
    at once.
    
    One thing that still needs to be added here is parallelization, since each
    of the folds can be calculated simultaneously.

    Parameters
    ----------
    X : pandas dataframe
        Feature or design marix.
    Y : pandas series
        Response or Target variable.
    cv_iterations : int
        The number of cross validation iterations. For k-fold, it is the
        number of folds, and for shuffle, it is the number of iterations.
    model : sklearn, or similar
        model to fit to the data. Must have fit, predict, and/or predict_proba
        methods, depending on the metrics.
    method : str, optional
        The method of cross-validation, either k-fold or shuffle.
        The default is 'k-fold'.
    mtype : str, optional
        The type of model, currently either classification or regressor.
        The default is 'classification'.
    stratified : bool, optional
        Use a stratified sample instead of a random sample.
        The default is True.
    print_ : bool, optional
        Print the results along the way. The default is True.
    random_state : int, optional
        Set the random seed for reproducibility. The default is None.
    method_on_X : function, optional
        An optional function to pass that performs an operation on X.
        For example, you could pass a function to perform PCA before
        performing the fits. This would enable the function to be applied to
        each level of split. The method used to transform/modify X can be
        of any class, the only requirement is that it has both fit and
        transform methods, and that the arguments for fit method accept both
        an X and Y argument, even if the Y argument does nothing. You may need
        to create a simple wrapper for this, if you have a known method you 
        want to use, but doesn't quite fit what you need. The default is None.
    mox_args : dict, optional
        Any optional arguments that get passed to the constructor of the
        method_on_X. The default is {}.
    avg : str, optional
        for metrics where it applies, such as AUC, how to calculate the
        metric (see sklearn docs for more details, particularily with AUC).
        The default is "weighted".
    shuffle : bool, optional
        Shuffle the dataset in the k-fold operations. The default is True.
    test_size : float, optional
        A percent representing the size of the test set. The default is 0.1.
    calculate : list or str, optional
        What to calculate the metrics on, either in-sample or out-of-sample,
        or both. In-sample measures the metrics on the train set while
        out-of-sample measures on the test set.
        The default is ['Out of Sample'].
    metrics : list, optional
        The metrics to include in the analysis.
        The default is ['Accuracy', 'F1', 'Sens/Recall',
                        'Specificity', 'ppv', 'npv', 'AUC'].
    Y_for_test_only : pandas series or numpy array, optional
        An alternate target variable to test on, mainly for the use case of
        training on one response (perhaps one that is more informative or
        restrictive), but then predicting on a seperate test set.
        The default is None.
    sample_limit : int, optional
        The minimum acceptable samples in a split for the response
        variable. Only applies to Y_for_test_only currently.
        The default is 20.

    Returns
    -------
    scores : dict
        Dictionary of performance metrics.

    '''
    splits = None
    scores = None
    if mtype != 'classification':
        stratified = False
        
    if method == 'k-fold':
        if stratified:
            splits = StratifiedKFold(cv_iterations, shuffle=shuffle,
                                     random_state=random_state)
        else:
            splits = KFold(cv_iterations, shuffle=shuffle,
                           random_state=random_state)
    else:
        if stratified:
            splits = StratifiedShuffleSplit(cv_iterations, test_size=test_size,
                                            random_state=random_state)
        else:
            splits = ShuffleSplit(cv_iterations, test_size=test_size,
                                  random_state=random_state)

    scores = runScorers(X, Y, splits, model, mtype, metrics=metrics, avg=avg,
                        calculate=calculate, method_on_X=method_on_X,
                        mox_args=mox_args, Y_for_test_only=Y_for_test_only,
                        sample_limit=sample_limit)

    if print_:
        printScores(scores)
            
    return scores


# labels is a dictionary identifying what the label values should be,
# where the dictionary key is the new label, and the dictionary value
# is the current class value
def prettyConfMat(Ytrue, Ypred,
                  print_=True, margins=True, labels=None):
    '''
    Print a pretty confusion matrix.

    Parameters
    ----------
    Ytrue : numpy array or pandas series
        The true class values.
    Ypred : numpy array or pandas series
        The predicted class values.
    print_ : bool, optional
        Print the confusion matrix at the end. The default is True.
    margins : bool, optional
        Include the summed margins in the matrix. The default is True.
    labels : dict, optional
        A dictionary of key-value pairs, where the keys are the labels,
        and the values are the class values in Ytrue. class labels for the
        confusion matrix. The default is None.

    Raises
    ------
    KeyError
        If labels are provided, an error is raised if the size of the array
        containing the labels does not match the number of possible classes.

    Returns
    -------
    confmat : pandas dataframe
        A dataframe containing the confusion matrix.

    '''
    if labels is not None:
        if isinstance(Ytrue, pd.DataFrame) or isinstance(Ytrue, pd.Series):
            Ytrue = Ytrue.copy().values
        if isinstance(Ypred, pd.DataFrame) or isinstance(Ypred, pd.Series):
            Ypred = Ypred.copy().values
        maxlen = np.unique(Ytrue).shape[0]
        if len(labels) != maxlen:
            raise KeyError('Incorrect number of labels compared to truth label values')
        else:
            idx = np.arange(0, Ytrue.shape[0])
            Ytrue_label = pd.Series(index=idx, dtype=object)
            Ypred_label = pd.Series(index=idx, dtype=object)
            for lab in labels.keys():
                Ytrue_label.loc[Ytrue == labels[lab]] = lab
                Ypred_label.loc[Ypred == labels[lab]] = lab
    else:
        Ytrue_label = Ytrue.copy()
        Ypred_label = Ypred.copy()
    
    cm_vals = pd.crosstab(Ytrue_label, Ypred_label, margins=margins, margins_name='Total')

    rowcolkeys = [lab for lab in labels.keys()] + ['Total']
    cm_vals = cm_vals.reindex(rowcolkeys, axis=0)
    cm_vals = cm_vals.reindex(rowcolkeys, axis=1)
    
    rownames = pd.MultiIndex.from_product([['Actual'], rowcolkeys])
    colnames = pd.MultiIndex.from_product([['Predicted'], rowcolkeys])
    #confmat = cm_vals.reindex(pd.MultiIndex.from_product([['Actual'], [lab for lab in labels.keys()] + ['All']]))
    confmat = pd.DataFrame(cm_vals.values, index=rownames, columns=colnames)

    if print_:
        print(confmat)
    return confmat


def RegressionSE(X, Y, fit_mod, logit=True, low_memory=False):
    '''
    Since sklearn doesn't generate the standard errors by default...

    Parameters
    ----------
    X : pandas dataframe
        The feature or design matrix.
    Y : pandas series or numpy array
        The response or target variable.
    fit_mod : sklearn model, or similar
        The fitted model (Logistic or linear Regression).
    logit : bool, optional
        If True, calculate Logistic Regression SE, otherwise, Linear. Default is True
    low_memory : bool, optional
        Sometimes this process uses a lot of memory. If True, calculate using less memory.
        The default is False.

    Returns
    -------
    se : numpy array
        The standard error of the coefficients.

    '''
    se = np.nan
    Xt = np.hstack([np.ones((X.shape[0], 1)), X])
    if logit:
        #Ypred = fit_mod.predict(X)
        Yprob = fit_mod.predict_proba(X)
        Yprob_prod = np.product(Yprob, axis=1)
        if low_memory:
            XV = np.zeros(shape=Xt.T.shape)
            for j in range(X.T.shape[1]):
                XV[:, j] = Xt.T[:, j] * Yprob_prod[j]
            XVXt = np.zeros((XV.shape[0], XV.shape[0]))
            try:
                XVXt = XV.dot(Xt)
            except MemoryError:
                for i in range(XV.shape[0]):
                    for k in range(XV.shape[1]):
                        sumprod = np.sum(np.product(XV[i, :], Xt[:, k]))
                        XVXt[i, k] = sumprod
        else:
            # p*(1-p), or pq
            Yprob_prod = np.diagflat(np.product(Yprob, axis=1))
            # cov of logit
            XVXt = np.dot(np.dot(Xt.T, Yprob_prod), Xt)
        cov = np.linalg.inv(XVXt)
    else:
        betas = np.concatenate((fit_mod.intercept_, fit_mod.coef_)).reshape(-1, 1)
        sigma2 = np.sum((Y - Xt.dot(betas))**2)/(Xt.shape[0] - Xt.shape[1])
        cov = sigma2*np.linalg.inv(np.dot(Xt.T, Xt))
    se = np.sqrt(np.diag(cov))  
    return se


def corrmat_validity_check(corr_mat, corr_tol=1e-8):
    '''
    Check for perfect correlation and/or a singular covariance matrix. Either
    of these conditions could indicate a matrix unsuitable for Linear
    Regression.

    Parameters
    ----------
    corr_mat : numpy array
        The correlation matrix.
    corr_tol : float
        The floating point error allowed to check for perfect correlation.

    Returns
    -------
    has_perfect_corr : numpy array
        An array of boolean values indicating perfect correlation.

    '''
    det = np.linalg.det(corr_mat)
    ncols = corr_mat.shape[0]
    corr_mat_flat = corr_mat.reshape(-1, )
    has_perfect_corr = (np.abs(corr_mat_flat) >= 1-corr_tol).reshape(ncols, ncols)
    np.fill_diagonal(has_perfect_corr, 0)
    return has_perfect_corr, det


def vif(data, root=False, corr_tol=1e-8, sing_tol=1e-15):
    '''
    Calculate Variance Inflation Factors. The sqrt of vif indicates how many
    times larger the standard error is than it would be if that variable had
    no correlation with the other variables 
    
    Assumes rows are observations and columns are variables.
    
    Expects pandas.DataFrame or numpy.array

    Parameters
    ----------
    data : pandas dataframe or numpy array
        The feture or design matrix.
    root : bool, optional
        Return the sqrt of the variance inflation factors. The default is False.

    Returns
    -------
    Either pandas Dataframe or numpy array
        The VIFs.

    '''
    corr = np.corrcoef(data, rowvar=False)
    perfect_corr, det = corrmat_validity_check(corr, corr_tol=corr_tol)
    if perfect_corr.sum() > 0:
        warnings.warn("""There are perfectly correlated variables. VIF
                         may not be reliable.""")
    if det < sing_tol:
        warnings.warn("""The corr_mat determinant is essentially zero. VIF
                         may not be reliable.""")
    
    vifs = np.diag(np.linalg.inv(corr))
    
    if isinstance(data, pd.DataFrame):
        vifs = pd.DataFrame(vifs,
                            index=data.columns,
                            columns=['Variance Inflation Factors'])
    if np.any(vifs < 1):
        warnings.warn("""Some vifs are less than one. Model is not correctly
                      specified or is not a suitable linear model""")
                      
    if root:
        return np.sqrt(vifs)
    else:
        return vifs


def beta_trans(coef, type_="percent"):
    '''
    Transform the model coefficients in a logistic regression model.

    Parameters
    ----------
    coef : numpy array
        The model coefficients.
    type_ : str, optional
        The type of transform, either percent, odds, or decimal.
        
        Decimal and percent are the same, except that percent = decimal*100
        
        The default is "percent".

    Returns
    -------
    numpy array
        Transformed Coefficients.

    '''
    val = np.exp(coef)
    if type_ == "odds":
        return val
    elif type == "decimal":
        return val - 1
    elif type_ == "percent":
        return 100.0*(val - 1)
    else:
        return coef
    

def getPriors(data, Ycol):
    '''
    Calculate the class distributions across each column of the dataset.
    
    For classification problems, but could easily be used with regression
    problems if an idicator variable was made that indicated some split in the
    data.

    Parameters
    ----------
    data : pandas dataframe
        The feature or design matrix, along with the reponse variable.
    Ycol : str
        The column representing the target or response variable.

    Returns
    -------
    alltab : pandas dataframe
        A dataframe containing all of the results.

    '''
    tab = None
    alltab = None
    #numeric = False
    sub = data
    for k, i in enumerate(sub.columns[:-1]):
        if len(sub.loc[:, i].unique()) > 7:
            arr = np.array([[sub[sub[Ycol] == 0][i].mean(),
                             sub[sub[Ycol] == 1][i].mean(),
                             sub.loc[:,i].mean()],
                            [sub[sub[Ycol] == 0][i].std(),
                             sub[sub[Ycol] == 1][i].std(),
                             sub.loc[:,i].std()]])
            tab = pd.DataFrame(arr, index=["Mean", "Std. Dev"], columns=[0, 1, u'All'])
            tab.columns.name = Ycol
            tab.index.name = i
            tab.index = pd.MultiIndex.from_arrays([[tab.index.name]*len(tab.index), tab.index], names=(None, None))
        else:
            tab = pd.crosstab(sub.loc[:, i], sub.loc[:, Ycol], margins=True, normalize=True, dropna=False)
            tab.index = pd.MultiIndex.from_arrays([[tab.index.name]*len(tab.index), tab.index], names=(None, None))
        if k == 0:
            alltab = tab
        else:
            alltab = pd.concat((alltab, tab))
    return alltab
    

class EstimatorSelectionHelper(object):
    '''
    This is adapted from David Batista, so credit goes to him. 
    
    see here: davidsbatista.net/blog/2018/02/23/model_optimization
    
    I added some functionality that will extend to the sklearn-deap package
    as well. see here: https://github.com/rsteca/sklearn-deap
    '''
    def __init__(self, models, params, searchCV, searchCVparams):
        '''
        Constructor for the EstimatorSelectionHelper class.

        Parameters
        ----------
        models : dict
            A dictionary of model types.
        params : dict
            A dictionary of parameters for each model.
        searchCV : sklearn object
            The type of search being performed.
        searchCVparams : dict
            A dictionary of optional arguments to send to searchCV.

        Raises
        ------
        ValueError
            Raised when a model is missing parameters, or there is a mismatch.

        Returns
        -------
        None.

        '''
        if not set(models.keys()).issubset(set(params.keys())):
            missing_params = list(set(models.keys()) - set(params.keys()))
            raise ValueError("Some estimators are missing parameters: %s" % missing_params)
        self.models = models
        self.params = params
        self.keys = models.keys()
        self.grid_searches = {}
        self.searchCV = searchCV
        self.searchCVparams = searchCVparams
        self.compute_times = {}
        self.gs_name = ""
        
    def fit(self, X, y):
        '''
        perform the search.

        Parameters
        ----------
        X : pandas dataframe or numpy array
            The feature or design matrix.
        y : pandas series or numpy array
            The response or target variable.

        Returns
        -------
        self
            Adds fitted parameters/models to the object.

        '''
        for key in self.keys:
            t0 = dt()
            print("Running GridSearch for %s." % key)
            model = self.models[key]
            params = self.params[key]
            gs = self.searchCV(model, params, **self.searchCVparams)
            gs.fit(X, y)
            self.grid_searches[key] = gs
            self.compute_times[key] = dt() - t0
            self.gs_name = gs.__class__.__name__
        return self
    
    def _score_summary_ea(self, sort_by):
        '''
        Create a summary of the evolutionary algorithm results.

        Parameters
        ----------
        sort_by : list or list-like
            Columns to sort by, for example, 'mean_score'.

        Returns
        -------
        pandas dataframe
            A dataframe containing the results.

        '''
        rows = []
        for k in self.grid_searches:
            params = self.grid_searches[k].cv_results_['params']
            min_test_score = self.grid_searches[k].cv_results_["min_test_score"]
            max_test_score = self.grid_searches[k].cv_results_["max_test_score"]
            mean_test_score = self.grid_searches[k].cv_results_["mean_test_score"]
            std_test_score = self.grid_searches[k].cv_results_["std_test_score"]

            for i in range(len(min_test_score)):
                d = {
                    'estimator': k,
                    'min_score': min_test_score[i],
                    'max_score': max_test_score[i],
                    'mean_score': mean_test_score[i],
                    'std_score': std_test_score[i],
                    'all_params': params,
                }
                new_row = pd.Series({**params[i], **d})
                rows.append((new_row))
        
        df = pd.concat(rows, axis=1).T.sort_values([sort_by], ascending=False)
        
        columns = ['estimator', 'min_score', 'mean_score', 'max_score', 'std_score']
        columns = columns + [c for c in df.columns if c not in columns]
        
        return df[columns]
    
    def _score_summary_skgs(self, sort_by):
        '''
        Create a summary of the Grid Search results.

        Parameters
        ----------
        sort_by : list or list-like
            Columns to sort by, for example, 'mean_score'.

        Returns
        -------
        pandas dataframe
            A dataframe containing the results.

        '''
        def row(key, scores, params, score_type):
            d = {
                'estimator': key,
                'min_score': min(scores),
                'max_score': max(scores),
                'mean_score': np.mean(scores),
                'std_score': np.std(scores),
                'score_type': score_type
            }
            return pd.Series({**params, **d})
        
        rows = []
        for k in self.grid_searches:
            params = self.grid_searches[k].cv_results_['params']
            for score_method_key in self.searchCVparams['scoring'].keys():
                scores = []
                for i in range(self.grid_searches[k].cv):
                    key = "split{}_test_{}".format(i, score_method_key)
                    r = self.grid_searches[k].cv_results_[key]
                    scores.append(r.reshape(len(params), 1))
            
                all_scores = np.hstack(scores)
                for p, s in zip(params, all_scores):
                    rows.append((row(k, s ,p, score_method_key)))
        
        df = pd.concat(rows, axis=1).T.sort_values([sort_by], ascending=False)
        
        columns = ['estimator', 'min_score', 'mean_score', 'max_score', 'std_score']
        columns = columns + [c for c in df.columns if c not in columns]
        
        return df[columns]
    
    def score_summary(self, sort_by='mean_score'):
        '''
        Create a summary of the scores and results.

        Parameters
        ----------
        sort_by : list or list-like
            Columns to sort by, for example, 'mean_score'.

        Returns
        -------
        scores : pandas dataframe
            A dataframe containing the results.

        '''
        scores = None
        if self.gs_name == 'EvolutionaryAlgorithmSearchCV':
            scores = self._score_summary_ea(sort_by=sort_by)
        else:
            scores = self._score_summary_skgs(sort_by=sort_by)
        return scores


def BPCA_initmodel(y, q):
    '''
    Initialize the bPCA model.

    Parameters
    ----------
    y : numpy array
        The data to be nan-filled.
    q : int
        The number of dimensions to consider in the PCA.

    Returns
    -------
    M : dict
        A dictionary of the initialized values.

    '''
    M = {}
    M['N'] = y.shape[0]
    M['d'] = y.shape[1]
    M['q'] = q
    M['yest'] = y.copy()
    M['missidx'] = []
    M['nomissidx'] = []
    M['gnomiss'] = []
    M['gmiss'] = []
    
    for i in range(y.shape[0]):
        M['missidx'].append(np.where(np.isnan(y[i, :]))[0])
        M['nomissidx'].append(np.where(~np.isnan(y[i, :]))[0])
        if M['missidx'][i].shape[0] == 0:
            M['gnomiss'].append(i)
        else:
            M['gmiss'].append(i)
            M['yest'][i, M['missidx'][i]] = 0
    
    # ynomiss = y[M['gnomiss'], :]
    covy = np.cov(M['yest'], rowvar=False)
    U, S, _ = np.linalg.svd(covy, full_matrices=True)
    U = U[:, :q]
    S = S[:q]
    #V = V.T[:, :q]

    M['mu'] = np.nansum(y, axis=0)/np.array([np.sum(~np.isnan(y[:, col])) for col in range(y.shape[1])])

    M['W'] = U * np.sqrt(S);
    M['tau'] = 1/(np.sum(np.diag(covy)) - np.sum(np.diag(S)))
    taumax = 1e10
    taumin = 1e-10
    M['tau'] = max(min(M['tau'], taumax), taumin)
    
    M['galpha0'] = 1e-10
    M['balpha0'] = 1
    M['alpha'] = (2*M['galpha0'] + M['d'])/(M['tau']*np.diag(M['W'].T.dot(M['W'])) + 2*M['galpha0']/M['balpha0'])
    
    M['gmu0'] = 0.001
    
    M['btau0'] = 1
    M['gtau0'] = 1e-10
    M['SigW'] = np.eye(q)
    
    return M


def BPCA_dostep(M, y):
    '''
    The workhorse of the bPCA algorithm, the Expectation-Maximization
    step.

    Parameters
    ----------
    M : dict
        A dictionary containing preliminary results for the algorithm.
    y : numpy array
        The original data.

    Returns
    -------
    M : dict
        The updated results.

    '''
    N = M['N']
    d = M['d']
    
    Rx = np.eye(M['q']) + M['tau']*M['W'].T.dot(M['W']) + M['SigW']
    Rxinv_o = np.linalg.inv(Rx)

    idx = M['gnomiss']
    n = len(idx)
    
    dy = y[idx, :] - np.tile(M['mu'], (n, 1))
    x = M['tau'] * Rxinv_o.dot(M['W'].T).dot(dy.T)

    Tt = dy.T.dot(x.T)
    trS = np.sum(dy * dy)
    for i in M['gmiss']:
        dyo = y[i, M['nomissidx'][i]] - M['mu'][M['nomissidx'][i]]

        Wm = M['W'][M['missidx'][i], :]
        Wo = M['W'][M['nomissidx'][i], :]
        Rxinv = np.linalg.inv(Rx - M['tau']*Wm.T.dot(Wm))
        ex = M['tau']*Wo.T.dot(dyo.T)
        x = Rxinv.dot(ex)
        
        dym = Wm.dot(x)
        dy = y[i, :].copy()
        dy[M['nomissidx'][i]] = dyo.T
        dy[M['missidx'][i]] = dym.T
        M['yest'][i, :] = dy + M['mu']

        Tt = Tt + reshape_to_vect(dy).dot(reshape_to_vect(x).T)
        Tt[M['missidx'][i], :] = Tt[M['missidx'][i], :] + Wm.dot(Rxinv)
        trS = trS + dy.dot(dy.T) + len(M['missidx'][i])/M['tau'] + np.trace(Wm.dot(Rxinv).dot(Wm.T))

    Tt = Tt/N
    trS = trS/N

    Dw = Rxinv_o + M['tau']*Tt.T.dot(M['W']).dot(Rxinv_o) + np.diag(M['alpha'])/N
    Dwinv = np.linalg.inv(Dw)
    M['W'] = Tt.dot(Dwinv)
    
    M['tau'] = (d + 2*M['gtau0']/N)/(trS - np.trace(Tt.T.dot(M['W'])) + (M['mu'].dot(M['mu'].T)*M['gmu0'] + 2*M['gtau0']/M['btau0'])/N)
    M['SigW'] = Dwinv*(d/N)
    M['alpha'] = (2*M['galpha0'] + d)/(M['tau']*np.diag(M['W'].T.dot(M['W'])) + np.diag(M['SigW']) + 2*M['galpha0']/M['balpha0'])

    return M
        


# assumes no rows of all NaNs
def bPCA(data, k=None, maxepoch=200, stepsize=10, dtau_tol=1e-8):
    '''
    An algorithm to compute missing values using Bayesian PCA. Translated from
    MATLAB code by Shigeyuki OBA, 2002 May. 5th.

    Parameters
    ----------
    data : numpy array
        The data with missing values to be estimated.
    k : int, optional
        The number of components to consider. Must be less than the number of
        columns. If None, use num_cols - 1. The default is None.
    maxepoch : int, optional
        Number of iterations. The default is 200.
    stepsize : int, optional
        The number of iterations to compute before printing results.
        The default is 10.
    dtau_tol : float, optional
        The precision tolerance. Breaks if precision is lower than the
        tolerance. The default is 1e-8.

    Returns
    -------
    M : dict
        A dictionary of the results, containing:
            mu: The estimated mean row vector
            W: The estimated principal axis matrix
            tau: The estimated precision (inverse variance) of the residual 
                 error

    '''
    N, d = data.shape
    if k is None:
        k = d - 1
    
    M = BPCA_initmodel(data, k)
    tauold = 1000
    
    for epoch in range(maxepoch):
        M = BPCA_dostep(M, data)
        if epoch % stepsize == 0:
            tau = M['tau'];
            dtau = np.abs(np.log10(tau)-np.log10(tauold));
            print('epoch=%d, dtau=%g' % (epoch, dtau));
            if dtau < dtau_tol:
                break
            tauold = tau    
    return M
        


import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import os

from datetime import datetime
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import f1_score
from time import mktime
from timeit import default_timer as dt

from .feature_selection import dropHighCorrs, getHighCorrs, naiveVarDrop
from .metrics import aicCalc, glm_regularized_AIC as glmAIC
from .modeling import crossVal, prettyConfMat
from .parallel import p_prog_simp
from .utils import print_time, updateProgBar


def calcAICsLasso(penalty, Xtrain, Ytrain, sample_size, unreg_full_mod=None,
                  random_state=123):
    '''
    Utility function for quick_analysis. Need to update this to include more
    GLM models.

    Parameters
    ----------
    penalty : flost
        The strength of the L1 penalty.
    Xtrain : numpy array or pandas dataframe
        The design or feature matrix.
    Ytrain : numpy array or pandas series
        The target or response variable.
    sample_size : int
        The number of observations.
    unreg_full_mod : sklearn object, or similar, optional
        The unregularized full model. The default is None.
    random_state : int, optional
        Random seed for the process. The default is 123.

    Returns
    -------
    nzc : list
        A list indicating the non-zero coefficients.
    aics : list
        A list containing the AIC values.

    '''
    cur_mod_reg = LogisticRegression(C=penalty, max_iter=10000, penalty="l1",
                                     solver='liblinear',
                                     random_state=random_state,
                                     n_jobs=1).fit(Xtrain, Ytrain)
    Yprob = cur_mod_reg.predict_proba(Xtrain)[:, 1]
    Ypred = cur_mod_reg.predict(Xtrain)
    mod_coef = np.concatenate((cur_mod_reg.intercept_,
                               np.squeeze(cur_mod_reg.coef_)))
    nzc = sum((1 if coef != 0 else 0 for coef in mod_coef))
    loglike = np.sum(Ytrain*np.log(Yprob) + (1 - Ytrain)*np.log(1 - Yprob))
    
    aics = []
    aics.append(aicCalc(loglike, nzc, sample_size, c=2, metric="aic"))
    aics.append(aicCalc(loglike, nzc, sample_size, c=2, metric="aicc"))
    aics.append(aicCalc(loglike, nzc, sample_size, c=2, metric="bic"))
    aics.append(aicCalc(loglike, nzc, sample_size, c=2, metric="ebic"))
    aics.append(np.mean(Ypred != Ytrain))
    aics.append(f1_score(Ytrain, Ypred))
    if unreg_full_mod is not None:
        aics.append(glmAIC(Xtrain, Ytrain, cur_mod_reg, unreg_full_mod,
                             tol=1e-8, method="kawano"))
        aics.append(glmAIC(Xtrain, Ytrain, cur_mod_reg, unreg_full_mod,
                             tol=1e-8, method="hastie"))
    
    return nzc, aics


def parallel_AIC_data_retriever(save_path, save_name, num_chunks, sample_size,
                                unreg_full_mod, random_state, target_var,
                                cur_pred_list, penalty):
    '''
    Utility function for QuickAnalysis. Not to be used externally.

    Parameters
    ----------
    save_path : str
        The directory path to save temporary files to.
    save_name : str
        The name of the temporary file to save.
    num_chunks : int
        The number of chunks for the pandas msgpacks.
    sample_size : int
        The number of observations.
    unreg_full_mod : sklearn object, or similar
        Unregularized full model.
    random_state : int
        Random seed for the process.
    target_var : str
        The name of the target variable.
    cur_pred_list : list
        The list of current predictors in the model.
    penalty : float
        The strength of the L1 penalty.

    Returns
    -------
    res : tuple
        Tuple of results (non-zero coefficients and aic values).

    '''
    # Note: Should probably update this to feather (as of 2019)
    # since it is faster and better on memory in general.
    # Since this is not long term storage, feather would be a 
    # good choice. However, for long-term storage, parquet is 
    # probably the best option
    
    #data = pd.read_msgpack(save_path + save_name + '.msg')
    #chunks = [int(x[6:]) for x in data.keys()]
    #chunks = sorted(chunks)
    #mergedDF = pd.DataFrame()
    #for chunk in chunks:
    #    mergedDF = mergedDF.append(data['chunk_'+str(chunk)])
    mergedDF = pd.read_feather(save_path + save_name + '.fth')
    
    Xtrain = mergedDF.loc[:, cur_pred_list]
    Ytrain = mergedDF.loc[:, target_var]
    
    res = calcAICsLasso(penalty=penalty,
                        Xtrain=Xtrain,
                        Ytrain=Ytrain,
                        sample_size=sample_size,
                        unreg_full_mod=unreg_full_mod,
                        random_state=random_state)
    
    # For debugging purposes, to attempt to force garbage collection in parallel
    mergedDF = []
    Xtrain = []
    Ytrain = []
    return res


class QuickAnalysis(object):
    '''
    A class for performing a quick sweep of the data. Helps establish a
    baseline model to get started with, as well as quickly identify potentially
    useful features.
    '''
    def __init__(self, train, holdout, target_var, low_memory=False,
                 name="QuickAnalysis"):
        '''
        Constructor for QuickAnalysis

        Parameters
        ----------
        train : pandas dataframe
            The training dataset.
        holdout : pandas dataframe
            The holdout set (for evaluation of the final model).
        target_var : str
            The target or response variable.
        low_memory : bool, optional
            For instances where it applies, attempts to use less memory.
            The default is False.
        name : str, optional
            A name for the modeling object. The default is "QuickAnalysis".

        Returns
        -------
        None.

        '''
        self._full_analysis_arglist = None
        self._bdp_analysis_arglist = None
        self._fsa_analysis_arglist = None
        self.name = name
        self.train_raw = train.copy()
        self.holdout_raw = holdout.copy()
        if low_memory:
            self.train_raw = None
            self.holdout_raw = None
        self.low_memory = low_memory
        self.train = train.copy()
        self.holdout = holdout.copy()
        self.target_var = target_var
        self.steps = {}
        self.step_num = 0
        self.predictors_raw = train.columns[np.where(target_var != train.columns)[0]]
        self.cur_pred_list = train.columns[np.where(target_var != train.columns)[0]]
        self.HCdropList = None
        self.y_corr = None
        self.corr_mat = None
        self.highCorr = None
        self.varDropList = None
        self.aics = None
        self.nonzero_coefs = None
        self.reduction = None
        self.opt_c = None
        self.optimal_num_params = None
        self.forwardElimMetrics = None
        self.num_params_forward = None
        self.coef_ord_red_final = None
        self.num_dropped_cols = {}
        
    
    def GenCorrStats(self, is_raw=False):
        '''
        Generate Correlation Statistics.

        Parameters
        ----------
        is_raw : bool, optional
            Is the raw data (vs. scaled data). The default is False.

        Returns
        -------
        None.

        '''
        corr_mat = None
        if is_raw:
            corr_mat = pd.DataFrame(np.corrcoef(self.train_raw.values,
                                                rowvar=False),
                                    index=self.train_raw.columns,
                                    columns=self.train_raw.columns)
        else:
            corr_mat = pd.DataFrame(np.corrcoef(self.train.values,
                                                rowvar=False),
                                    index=self.train.columns,
                                    columns=self.train.columns)
        self.y_corr = corr_mat.loc[self.cur_pred_list, self.target_var]
        self.corr_mat = corr_mat
    
    def sort_y_corr(self):
        '''
        Sort columns by their correlation with the target variable.

        Returns
        -------
        pandas series
            Sorted Correlations.

        '''
        if self.y_corr is None:
            print("No y_corr recorded. Run GenCorrMat")
        else:
            return self.y_corr.abs().sort_values(ascending=False)
    
    def plot_y_corr(self):
        '''
        Plot variable correlations with the target variable.

        Returns
        -------
        None.

        '''
        sorted_corrs_with_y = self.y_corr.sort_values()

        plt.figure(figsize=(10, 5))
        ax = plt.subplot(1, 1, 1)

        ax.plot(sorted_corrs_with_y.values)
        plt.ylabel("Correlation")
        plt.xlabel("Sorted Features")
        plt.title("Correlation of Each Feature with " + self.target_var)
        plt.show()
        sorted_corrs_with_y = []
    
    def genHighCorrs(self, corr_cutoff):
        '''
        Generate a list of variables that are highly correlated with one
        another.

        Parameters
        ----------
        corr_cutoff : float
            Cutoff value to determine high correlations.

        Returns
        -------
        None.

        '''
        self.highCorr = getHighCorrs(self.corr_mat, corr_cutoff)
    
    def calcAICs(self, num_cs, try_parallel, n_jobs, random_state, remove_msg,
                 chunk):
        '''
        Calculate AIC values for the models to prepare for comparison.

        Parameters
        ----------
        num_cs : int, optional
            number of penalty values to test (i.e. number of models to evaluate
            during AIC calculations). 
        try_parallel : bool
            Whether to try running the operation in parallel.
        n_jobs : int
            Number of parallel processes to run.
        random_state : int
            Random seed for the process.
        remove_msg : bool
            Remove the temporary pandas msgpacks once calculations are
            complete.
        chunk : bool
            Break the operation into chunks.

        Returns
        -------
        None.

        '''
        nonzero_coefs = []
        aics = np.zeros(shape=(num_cs, 6))
        
        res = None
        
        if try_parallel:
            num_chunks = int(((self.train.memory_usage(index=True).sum()/(1024**3))/2) + 1)
            
            save_path = 'qr_tmp'
            if not os.path.isdir(save_path):
                os.mkdir(save_path)
            save_name = '/qr_run_' + str(int(mktime(datetime.now().timetuple())))
            
            #pd.to_msgpack(save_path + save_name + '.msg', {
            #        'chunk_{0}'.format(i):chunk for i, chunk in enumerate(np.array_split(self.train, num_chunks))
            #                                              })
            temp_train = self.train.copy(deep=True)
            temp_train.to_feather(save_path + save_name + '.fth')
            
            size = self.train.shape[0]
            arr = {'save_path': save_path,
                   'save_name': save_name,
                   'num_chunks': num_chunks,
                   'sample_size': size,
                   'unreg_full_mod': None,
                   'random_state': random_state,
                   'target_var': self.target_var,
                   'cur_pred_list': list(self.cur_pred_list),
                  }
            
            if chunk:
                res = []
                num_chunks = int((num_cs/n_jobs) + 1)
                
                for chunk in range(num_chunks):
                    loop_arr = [{'penalty': red} for red in self.reduction[chunk*n_jobs:((chunk+1)*n_jobs)]]
                    #temp_jobs = len(loop_arr)
                    new_res = p_prog_simp(args=arr, loop_args=loop_arr,
                                  function=parallel_AIC_data_retriever,
                                  n_jobs=n_jobs)
                res = res + new_res
            else:
                
                loop_arr = [{'penalty': red} for red in self.reduction]

                res = p_prog_simp(args=arr, loop_args=loop_arr,
                                  function=parallel_AIC_data_retriever,
                                  n_jobs=n_jobs)

            if remove_msg:
                #file_ = save_path + save_name + '.msg'
                file_ = save_path + save_name + '.fth'
                try:
                    os.remove(file_)
                except FileNotFoundError:
                    print('Could Not find file ', file_, 'Continuing...')
            
            arr = []
            loop_arr = []
        
        else: 
            res = [calcAICsLasso(penalty=red,
                                 Xtrain=self.train.loc[:, self.cur_pred_list],
                                 Ytrain=self.train.loc[:, self.target_var],
                                 sample_size=self.train.shape[0],
                                 unreg_full_mod=None,
                                 random_state=random_state) for red in self.reduction]
        
        for i, re in enumerate(res):
            nonzero_coefs.append(re[0])
            aics[i, :] = re[1]

        self.aics = aics
        self.nonzero_coefs = list(nonzero_coefs)
    
    def plotAICs(self):
        '''
        Plot the AIC calculations.

        Returns
        -------
        None.

        '''
        fig, ax = plt.subplots(figsize=(16,10))
        ax.plot(self.nonzero_coefs, self.aics[:, 1])
        ax.plot(self.nonzero_coefs, self.aics[:, 2])
        ax.set_ylabel("AIC/BIC Value")
        ax.set_xlabel("Number of Model Parameters")
        ax2 = ax.twinx()
        ax2.plot(self.nonzero_coefs, self.aics[:, 4], color="purple")
        ax2.plot(self.nonzero_coefs, self.aics[:, 5], color="green")
        ax2.set_ylabel("F1 Score and Misclassification Rate")
        ax.legend(["aicc", "bic"])
        ax2.legend(["misclass", "f1"])
        plt.title("Number of Non-zero Parameters by Change in L1 Penalty")
        plt.show()
        
    def forwardElimMetricCalc(self, coef_ord_red, stride, random_state):
        '''
        Run a very basic forward elimination procedure, and calculate model 
        metrics.

        Parameters
        ----------
        coef_ord_red : list
            A list of model coefficients, ordered by importance.
        stride : int
            Window size of the coefficients to test in the forward selection
            procedure. For example, stride=1 is 1,2,3,4,..., stride=2 is
            1,3,5,7,...
        random_state : int
            Random seed for procedure.

        Returns
        -------
        None.

        '''
        num_params = np.arange(1, self.train.shape[1]-1, stride)
        # f1, acc, sens, spec, auc
        metrics = np.zeros(shape=(len(num_params), 10))
        
        t0 = dt()

        for i, parm in enumerate(num_params):
            cur_mod_reg = LogisticRegression(C=self.opt_c, max_iter=10000,
                                             penalty="l1", solver='liblinear',
                                             random_state=random_state)
            # Cross-Val fully reduced model
            cols_to_include = list(coef_ord_red.index[0:parm])
            cur_mod_cv_results = crossVal(self.train.loc[:, cols_to_include],
                                          self.train.loc[:, self.target_var],
                                          5, cur_mod_reg, print_=False)

            metrics[i, 0:2] = (np.mean(cur_mod_cv_results['Out of Sample']["Accuracy"]),
                               np.std(cur_mod_cv_results['Out of Sample']["Accuracy"], ddof=1))
            metrics[i, 2:4] = (np.mean(cur_mod_cv_results['Out of Sample']["F1"]),
                               np.std(cur_mod_cv_results['Out of Sample']["F1"], ddof=1))
            metrics[i, 4:6] = (np.mean(cur_mod_cv_results['Out of Sample']["Sens/Recall"]),
                               np.std(cur_mod_cv_results['Out of Sample']["Sens/Recall"], ddof=1))
            metrics[i, 6:8] = (np.mean(cur_mod_cv_results['Out of Sample']["Specificity"]),
                               np.std(cur_mod_cv_results['Out of Sample']["Specificity"], ddof=1))
            metrics[i, 8:10] = (np.mean(cur_mod_cv_results['Out of Sample']["AUC"]),
                               np.std(cur_mod_cv_results['Out of Sample']["AUC"], ddof=1))
            
            updateProgBar(i + 1, len(num_params), t0)
        self.forwardElimMetrics = metrics
        self.num_params_forward = num_params
        
    def plotForwardMetrics(self):
        '''
        Plot forward elimination metric results.

        Returns
        -------
        None.

        '''
        
        plt.figure(figsize=(10,7))
        
        acc_mean = self.forwardElimMetrics[:, 0].reshape(-1, 1)
        f1_mean = self.forwardElimMetrics[:, 2].reshape(-1, 1)
        sens_mean = self.forwardElimMetrics[:, 4].reshape(-1, 1)
        spec_mean = self.forwardElimMetrics[:, 6].reshape(-1, 1)
        auc_mean = self.forwardElimMetrics[:, 8].reshape(-1, 1)
        
        acc_bnds = self.forwardElimMetrics[:, 1].reshape(-1, 1).dot(np.array([-2, 2]).reshape(1, 2)) + acc_mean
        f1_bnds = self.forwardElimMetrics[:, 3].reshape(-1, 1).dot(np.array([-2, 2]).reshape(1, 2)) + f1_mean
        sens_bnds = self.forwardElimMetrics[:, 5].reshape(-1, 1).dot(np.array([-2, 2]).reshape(1, 2)) + sens_mean
        spec_bnds = self.forwardElimMetrics[:, 7].reshape(-1, 1).dot(np.array([-2, 2]).reshape(1, 2)) + spec_mean
        auc_bnds = self.forwardElimMetrics[:, 9].reshape(-1, 1).dot(np.array([-2, 2]).reshape(1, 2)) + auc_mean
       
        plt.plot(self.num_params_forward, acc_mean)
        plt.fill_between(self.num_params_forward,
                         acc_bnds[:, 0], acc_bnds[:, 1], alpha=0.5)
        plt.plot(self.num_params_forward, f1_mean)
        plt.fill_between(self.num_params_forward,
                         f1_bnds[:, 0], f1_bnds[:, 1], alpha=0.5)
        plt.plot(self.num_params_forward, sens_mean)
        plt.fill_between(self.num_params_forward,
                         sens_bnds[:, 0], sens_bnds[:, 1], alpha=0.5)
        plt.plot(self.num_params_forward, spec_mean)
        plt.fill_between(self.num_params_forward,
                         spec_bnds[:, 0], spec_bnds[:, 1], alpha=0.5)
        plt.plot(self.num_params_forward, auc_mean)
        plt.fill_between(self.num_params_forward,
                         auc_bnds[:, 0], auc_bnds[:, 1], alpha=0.5)
        
        plt.xlabel("Number of Model Features")
        plt.ylabel("Metric Values")
        plt.title("Change in Model Metrics as Features Increase")
        plt.legend(["acc", "f1", "sens", "spec", "auc"])
        plt.show()
    
    def confusion_matrix(self, print_=True):
        '''
        Generate a confusion matrix (for classification data).

        Parameters
        ----------
        print_ : bool, optional
            Print the confusion matrix to the console. The default is True.

        Returns
        -------
        confmat : pandas dataframe
            A dataframe containing the confusion matrix.

        '''
        confmat = None
        if self.Ytest is None:
            print("Run QuickAnalysis first")
        else:
            confmat = prettyConfMat(self.Ytest, self.Ytest_pred, print_=print_)
        if print_:
            print(confmat)
        return confmat
    
    def build_model(self, name, cv_iterations, penalty, random_state):
        '''
        A utility function for building the model at different steps of the 
        QuickAnalysis process.

        Parameters
        ----------
        name : str
            The step where the model is being built.
        cv_iterations : int
            Number of cross validation steps.
        penalty : float
            The strength of the L1 penalty.
        random_state : int
            Random seed of the process.

        Returns
        -------
        None.

        '''
        mod_crossval = crossVal(self.train.loc[:, self.cur_pred_list],
                                self.train.loc[:, self.target_var],
                                cv_iterations=cv_iterations,
                                model=LogisticRegression(C=penalty,
                                                         penalty="l1",
                                                         solver="liblinear",
                                                         max_iter=10000,
                                                         random_state=random_state),
                                print_=False)
        self.steps[name] = {
                'crossval': mod_crossval,
                'model': LogisticRegression(C=penalty,
                                            penalty="l1",
                                            solver="liblinear",
                                            max_iter=10000,
                                            random_state=random_state).fit(self.train.loc[:, self.cur_pred_list],
                                                                           self.train.loc[:, self.target_var]),
                'predictors': list(self.cur_pred_list)
            }
        
    
    def BaselineDropProcedures(self, freshStart=True, downsample=False,
                               dropCols=None, dropCor=True, corr_cutoff=0.9,
                               dropVar=True, dropVarTol=0.001,
                               cv_iterations=5,
                               random_state=123, default_penalty=0.01,
                               verbose=True, t0=None):
        '''
        Create Baseline model and begin dropping features from the model 
        based on given criteria.

        Parameters
        ----------
        freshStart : bool, optional
            Whether to use a previous result, or start from scrath.
            The default is True.
        downsample : bool, optional
            Whether or not to downsample the majority class.
            The default is False.
        dropCols : list, optional
            List of columns to initially drop. The default is None.
        dropCor : bool, optional
            Drop variables with high correlations. The default is True.
        corr_cutoff : float, optional
            Cutoff for removing high correlation variables. Anything below the
            cutoff is kept. The default is 0.9.
        dropVar : bool, optional
            Drop variables with very low variance.
            Does nothing if data is standardized. The default is True.
        dropVarTol : float, optional
            Cutoff value for low variance. The default is 0.001.
        cv_iterations : int, optional
            The number of cross validation steps. The default is 5.
        random_state : int, optional
            The random seed for the process. The default is 123.
        default_penalty : float, optional
            The penalty term to be applied to models where the penalty is
            otherwise unspecified. The default is 0.01.
        verbose : bool, optional
            Print steps as they complete. The default is True.
        t0 : float, optional
            The start time of the process. For internal use.
            The default is None.

        Returns
        -------
        None.

        '''
        self._bdp_analysis_arglist = {'freshStart': freshStart,
                                 'downsample': downsample,
                                 'dropCols': dropCols,
                                 'dropCor': dropCor,
                                 'corr_cutoff': corr_cutoff,
                                 'dropVar': dropVar,
                                 'dropVarTol': dropVarTol,
                                 'cv_iterations': cv_iterations,
                                 'random_state': random_state,
                                 'default_penalty': default_penalty,
                                 'verbose': verbose,
                                 't0': t0
                                 }
        
        if t0 is None:
            t0 = dt()
        
        if freshStart:
            if not self.low_memory:
                self.train = self.train_raw
                self.holdout = self.holdout_raw
            else:
                print("Can't run fresh start in low memory mode.")
        
        if verbose:
            print_time("\nCreating Baseline Model...", t0, te=dt())
        
        self.build_model('baseline_mod', cv_iterations, default_penalty,
                         random_state)
        
        if verbose:
            print_time("\nExecuting Column Drop Procedures...", t0, te=dt())

        if dropCols is not None:
            if verbose:
                print_time("\nDropping Pre-Determined Columns...", t0, te=dt())

            cols = list(self.train.columns)
            # what I used to do... 
            #idx = list(np.where([True if sum((True if reject in col else False for reject in dropCols))>0 else False for col in cols])[0])
            idx = [i for i, col in zip(range(len(cols)),
                                       cols) if col in dropCols]
            label_idx = list(self.train.columns[idx])
            self.train = self.train.drop(label_idx, axis=1)
            self.cur_pred_list = self.train.columns[np.where(self.target_var != self.train.columns)[0]]
            self.num_dropped_cols['dropCol_list'] = len(dropCols)
            if verbose:
                print_time("\nDropped " + str(len(dropCols)) + " Predefined Columns...",
                           t0, te=dt())
        
        if dropVar:
            if verbose:
                print_time("\nDropping Low Variance Columns...", t0, te=dt())

            varDropList = naiveVarDrop(self.train,
                                       list(self.train.columns[np.where(self.target_var != self.train.columns)[0]]),
                                       tol=dropVarTol,
                                       asList=True)
            
            self.train = self.train.drop(varDropList, axis=1)
            self.varDropList = varDropList
            self.cur_pred_list = self.train.columns[np.where(self.target_var != self.train.columns)[0]]
            self.num_dropped_cols['varDropList'] = len(varDropList)
            if verbose:
                print_time("\nDropped " + str(len(varDropList)) + " Low-Variance Columns...", t0, te=dt())
            varDropList = []
        
        if dropCor:
            if verbose:
                print_time("\nGenerating/Recovering Correlation Matrix...", t0,
                           te=dt())

            if self.corr_mat is None or self.y_corr is None:
                self.GenCorrStats(is_raw=False)
            if self.highCorr is None:
                self.genHighCorrs(corr_cutoff)
        
            if verbose:
                print_time("\nDropping High Correlations...", t0, te=dt())

            HCdropList = dropHighCorrs(self.train, self.highCorr, asList=True,
                                       print_=False)
            self.train = self.train.drop(HCdropList, axis=1)
            self.HCdropList = HCdropList
            self.num_dropped_cols['HCdropList'] = len(HCdropList)
            if verbose:
                print_time("\nDropped " + str(len(HCdropList)) + " Correlated Columns...", t0, te=dt())
            HCdropList = []
            self.cur_pred_list = self.train.columns[np.where(self.target_var != self.train.columns)[0]]
            self.plot_y_corr()
        
        if verbose:
            total_dropped = 0
            for key in self.num_dropped_cols.keys():
                total_dropped += self.num_dropped_cols[key]
            print_time("\nDropped " + str(total_dropped) + " Columns...", t0,
                       te=dt())
        
        if verbose:
            print_time("\nCreating Post-Drop Baseline Model...", t0, te=dt())
        
        self.build_model('postdrop_baseline_mod', cv_iterations,
                         default_penalty, random_state)
        
        if verbose:
            print_time("\nFinished Baseline Drop...", t0, te=dt())
    
    def ForwardSelectionAnalysis(self, cv_iterations=5, random_state=123,
                                 default_penalty=0.01, for_stride=2,
                                 reduce_features_by=30, red_metric='bic',
                                 verbose=True, t0=None):
        '''
        Run very basic forward selection procedure on the model.

        Parameters
        ----------
        default_penalty : float, optional
            The penalty term to be applied to models where the penalty is
            otherwise unspecified. The default is 0.01.
        for_stride : int
            Window size of the coefficients to test in the forward selection
            procedure. For example, stride=1 is 1,2,3,4,..., stride=2 is
            1,3,5,7,... The default is 2.
        reduce_features_by : int, optional
            Number of features to evaluate in the final model.
            The default is 30.
        red_metric : str, optional
            Metric to use for finding optimal penalty value.
            The default is 'bic'.
        verbose : bool, optional
            Print results of the process. The default is True.
        t0 : float, optional
            Initial start time for the process. The default is None.

        Returns
        -------
        None.

        '''
        # TODO: make final feature selection automatic
        self._fsa_analysis_arglist = {
                                      'cv_iterations': cv_iterations,
                                      'random_state': random_state,
                                      'default_penalty': default_penalty,
                                      'for_stride': for_stride,
                                      'reduce_features_by': reduce_features_by,
                                      'red_metric': red_metric,
                                      'verbose': verbose,
                                      't0': t0
        }
        
        if t0 is None:
            t0 = dt()
        
        self.plotAICs()
        aic_type = {
            'aic': 0,
            'aicc': 1,
            'bic': 2,
            'ebic': 3,
            'misclass': 4
        }
        self.opt_c = self.reduction[np.where(self.aics[:, aic_type[red_metric]] == np.min(self.aics[:, aic_type[red_metric]]))][0]
        self.optimal_num_params = np.array(self.nonzero_coefs)[np.where(self.aics[:, aic_type[red_metric]] == np.min(self.aics[:, aic_type[red_metric]]))][0]
        
        if verbose:
            print_time("\nCreating Model with Optimal Penalty Value...", t0,
                       te=dt())
        
        self.build_model('lasso_reduction_mod', cv_iterations, self.opt_c,
                         random_state)
        
        coef = self.steps['lasso_reduction_mod']['model'].coef_.reshape(-1,)
        coef_ord = pd.DataFrame(np.abs(coef), index=self.train.loc[:, self.cur_pred_list].columns,
                                columns=["Importance"])
        coef_ord = coef_ord.sort_values('Importance', ascending=False)
        
        keep_cols = list(coef_ord[coef_ord["Importance"] > 0].index)
        keep_cols.append(self.target_var)
        self.train = self.train.loc[:, keep_cols]
        self.cur_pred_list = self.train.columns[np.where(self.target_var != self.train.columns)[0]]
        self.steps['lasso_reduction_mod']['keep_cols'] = keep_cols
        keep_cols = []
        coef_ord = []
        coef = []
        
        if verbose:
            print_time("\nCreating Model with Optimal Penalty Value After Removing Zeroed Parameters...\n",
                       t0, te=dt())
        
        self.build_model('lasso_reduction_mod_reduced', cv_iterations, 1,
                         random_state)
        
        coef_red = self.steps['lasso_reduction_mod_reduced']['model'].coef_.reshape(-1,1)
        coef_ord_red = pd.DataFrame(np.concatenate((np.abs(coef_red), coef_red,
                                                    100.0*(np.exp(coef_red)-1)),
                                                   axis=1),
                                    index=self.train.loc[:, self.cur_pred_list].columns,
                                    columns=["Importance",
                                             "Coefficiants",
                                             "% increase in Prob"])
        coef_ord_red = coef_ord_red.sort_values('Importance', ascending=False)
        
        self.steps['lasso_reduction_mod_reduced']['coef_ord_red'] = coef_ord_red
        
        self.forwardElimMetricCalc(coef_ord_red, for_stride, random_state)
        self.plotForwardMetrics()
        
        cols_to_include = list(coef_ord_red.index[:reduce_features_by])
        self.cur_pred_list = cols_to_include
        
        if verbose:
            print_time("\nCreating Final Model...", t0, te=dt())
        
        self.build_model('final_mod', cv_iterations, 1, random_state)
        
        coef_red_final = self.steps['final_mod']['model'].coef_.reshape(-1,1)
        coef_ord_red_final = pd.DataFrame(np.concatenate((np.abs(coef_red_final), coef_red_final,
                                                          100.0*(np.exp(coef_red_final)-1)), axis=1),
                                    index=self.train.loc[:, cols_to_include].columns,
                                          columns=["Importance", "Coefficiants", "% increase in Prob"])
        coef_ord_red_final = coef_ord_red_final.sort_values('Importance', ascending=False)
        self.coef_ord_red_final = coef_ord_red_final
        
        self.Xtrain = self.train.loc[:, cols_to_include]
        self.Ytrain = self.train.loc[:, self.target_var]
        self.Ypred_train = self.steps['final_mod']['model'].predict(self.Xtrain)
        self.Yprob_train = self.steps['final_mod']['model'].predict_proba(self.Xtrain)

        self.Xtest = self.holdout.loc[:, cols_to_include]
        self.Ytest = self.holdout.loc[:, self.target_var]
        self.Ypred_test = self.steps['final_mod']['model'].predict(self.Xtest)
        self.Yprob_test = self.steps['final_mod']['model'].predict_proba(self.Xtest)[:, 1]
        
        if verbose:
            print_time("\nFinished...", t0, te=dt())
        
        
    def RunFullAnalysis(self, freshStart=True, downsample=False, dropCols=None,
                        dropCor=True, corr_cutoff=0.9, dropVar=True,
                        dropVarTol=0.001, cv_iterations=5,
                        random_state=123, default_penalty=0.01, verbose=True,
                        try_parallel=True, for_stride=2, reduce_features_by=30,
                        num_cs=100, red_metric='bic', red_log_low=-5,
                        red_log_high=1, n_jobs=1, remove_msg=True,
                        chunk=False):
        '''
        Performs a quick analysis by first considering which features/variables
        may be uninformative predictors based on preliminary models. Then,
        it performs a very basic forward selection procedure to limit
        predictors in the final model. Currently, this only supports
        Logistic Regression.

        Parameters
        ----------
        freshStart : bool, optional
            Whether to use a previous result, or start from scrath.
            The default is True.
        downsample : bool, optional
            Whether or not to downsample the majority class.
            The default is False.
        dropCols : list, optional
            List of columns to initially drop. The default is None.
        dropCor : bool, optional
            Drop variables with high correlations. The default is True.
        corr_cutoff : float, optional
            Cutoff for removing high correlation variables. Anything below the
            cutoff is kept. The default is 0.9.
        dropVar : bool, optional
            Drop variables with very low variance.
            Does nothing if data is standardized. The default is True.
        dropVarTol : float, optional
            Cutoff value for low variance. The default is 0.001.
        cv_iterations : int, optional
            The number of cross validation steps. The default is 5.
        random_state : int, optional
            The random seed for the process. The default is 123.
        default_penalty : float, optional
            The penalty term to be applied to models where the penalty is
            otherwise unspecified. The default is 0.01.
        verbose : bool, optional
            Print steps as they complete. The default is True.
        try_parallel : bool, optional
            Try to perform the operation in parallel. If it fails, it will
            continue without parallel operations. The default is True.
        for_stride : int, optional
            The gap between variables evaluated for the forward selection
            metric. The default is 2.
        reduce_features_by : int, optional
            The size of the final model. The default is 30.
        num_cs : int, optional
            number of penalty values to test (i.e. number of models to evaluate
            during AIC calculations). The default is 100.
        red_metric : str, optional
            The metric used to optimize model reductions. The default is 'bic'.
        red_log_low : float, optional
            The logspace minimum value for evaluating the L1 penalty.
            The default is -5.
        red_log_high : float, optional
            The logspace maximum value for evaluating the L1 penalty.. The default is 1.
        n_jobs : int, optional
            The number of processes to attempt. The default is 1.
        remove_msg : bool, optional
            Remove messagepacks. The default is True.
        chunk : bool, optional
            Chunk msgpacks? The default is False.

        Returns
        -------
        None.

        '''
        t0 = dt()
        self._full_analysis_arglist = {'freshStart': freshStart,
                                 'downsample': downsample,
                                 'dropCols': dropCols,
                                 'dropCor': dropCor,
                                 'corr_cutoff': corr_cutoff,
                                 'dropVar': dropVar,
                                 'dropVarTol': dropVarTol,
                                 'cv_iterations': cv_iterations,
                                 'random_state': random_state,
                                 'default_penalty': default_penalty,
                                 'try_parallel': try_parallel,
                                 'for_stride': for_stride,
                                 'reduce_features_by': reduce_features_by,
                                 'num_cs': num_cs,
                                 'red_metric': red_metric,
                                 'red_log_low': red_log_low,
                                 'red_log_high': red_log_high,
                                 'n_jobs': n_jobs,
                                 'verbose': verbose,
                                 'remove_msg': remove_msg}
        
        self.BaselineDropProcedures(freshStart=freshStart,
                                    downsample=downsample,
                                    dropCols=dropCols,
                                    dropCor=dropCor,
                                    corr_cutoff=corr_cutoff,
                                    dropVar=dropVar,
                                    dropVarTol=dropVarTol,
                                    cv_iterations=cv_iterations,
                                    random_state=random_state,
                                    default_penalty=default_penalty,
                                    verbose=verbose,
                                    t0=t0)
        
        if verbose:
            print_time("\nCreating Lasso Reduction Models...", t0, te=dt())

        self.reduction = np.logspace(red_log_low, red_log_high, num_cs)
        if try_parallel:
            try_again = True
            while(try_again):
                try:
                    self.calcAICs(num_cs, try_parallel, n_jobs,
                                  random_state=random_state,
                                  remove_msg=remove_msg, chunk=chunk)
                    try_again = False
                except (MemoryError):
                    n_jobs = int(n_jobs - (n_jobs/2))
                    print('\n\tReduced n_jobs to ', n_jobs)
                    if n_jobs == 1:
                        try_again = False
                    if not try_again:
                        self.calcAICs(num_cs, False, n_jobs=1,
                                      random_state=random_state,
                                      remove_msg=remove_msg, chunk=chunk)
        else:
            self.calcAICs(num_cs, False, n_jobs=1, random_state=random_state,
                          remove_msg=remove_msg, chunk=chunk)
                
        self.ForwardSelectionAnalysis(cv_iterations=cv_iterations,
                                      random_state=random_state,
                                      default_penalty=default_penalty,
                                      for_stride=for_stride,
                                      reduce_features_by=reduce_features_by,
                                      red_metric=red_metric,
                                      verbose=verbose, t0=t0)



from concurrent.futures import (ProcessPoolExecutor, ThreadPoolExecutor,
                                as_completed)
from timeit import default_timer as dt
from .utils import updateProgBar


def p_prog_simp(args, loop_args, function, n_jobs=2, verbose=True,
                use_threads=True):
    # TODO: provide a fix that will work more consistently
    '''
    Parallel Progress bar.

    Note: it is a known bug for this function to fail while using sypder.
    I have not been able to discover why, or provide a fix. If you run a script
    with this function from the console (or possibly jupyter), it should run
    just fine. 
    
    It is also known that using threads may be buggy for this function. It
    was intended to be used only on processes, but I included the thread option
    as a supplement to experiment with later. It is not guarenteed to work.
    
    Parameters
    ----------
    args : dict
        Dictionary of stationary arguments (i.e. the arguments that are
        constant throughout the loop).
    loop_args : list
        A list of dictionaries that contain the arguments that may change at 
        any point during the loop (this is the iterator, and any additional
        arguments that may have different values at different stages of the
        loop).
    function : function
        The function being being processed in the loop (args and loop args will
        be passed to this function).
    n_jobs : int, optional
        The number of parallel processes to run. The default is 2.

    Returns
    -------
    list
        Returns a list of the function outputs for each iteration.

    '''
    t0 = dt()
    niter = len(loop_args)
    
    if n_jobs == 1:
        return_list = []
        for i, l_arg in enumerate(loop_args):
            return_list.append(function(**args, **l_arg))
            updateProgBar(i+1, niter, t0)
        return return_list
    
    if use_threads:
        pool = ThreadPoolExecutor(max_workers=n_jobs)
    else:
        pool = ProcessPoolExecutor(max_workers=n_jobs)
    with pool:
        futures = []
        for l_arg in loop_args:
            args.update(l_arg)
            futures.append(pool.submit(function,  **args))
        if verbose:
            for k, f in enumerate(as_completed(futures)):
                updateProgBar(k+1, niter, t0)
    out = []
    for i, future in enumerate(futures):
        try:
            out.append(future.result())
        except Exception as e:
            out.append(e)
    return out


def p_prog_simp_memorySafe(args, loop_args, function, n_jobs=2):
    '''
    Parallel Progress bar. The same as p_prog_simp, except it is more explicit
    in it's definition. This function was developed to try and limit
    memory consumption for parallel operations. I'm not 100% sure if it
    accomplishes that... More testing is needed.

    Parameters
    ----------
    args : dict
        Dictionary of stationary arguments (i.e. the arguments that are
        constant throughout the loop).
    loop_args : list
        A list of dictionaries that contain the arguments that may change at 
        any point during the loop (this is the iterator, and any additional
        arguments that may have different values at different stages of the
        loop).
    function : function
        The function being being processed in the loop (args and loop args will
        be passed to this function).
    n_jobs : int, optional
        The number of parallel processes to run. The default is 2.

    Returns
    -------
    list
        Returns a list of the function outputs for each iteration.

    '''
    t0 = dt()
    niter = len(loop_args)
    
    if n_jobs == 1:
        return_list = []
        for i, l_arg in enumerate(loop_args):
            return_list.append(function(**args, **l_arg))
            updateProgBar(i+1, niter, t0)
        return return_list
        
    with ProcessPoolExecutor(max_workers=n_jobs) as pool:
        current_futures = []
        out = []
        jobs_complete = 0
        jobs_assigned = 0
        next_job = 0
        while (len(out) < niter):
            if len(current_futures) < n_jobs:
                if jobs_assigned < niter:
                    args.update(loop_args[next_job])
                    current_futures.append(pool.submit(function,  **args))
                    next_job += 1
                    jobs_assigned += 1
            for spot, fut in enumerate(current_futures):
                if fut.done():
                    try:
                        out.append(fut.result())
                    except Exception as e:
                        out.append(e)
                    del current_futures[spot]
                    jobs_complete += 1
                    updateProgBar(jobs_complete, niter, t0)
    return out
